"""This module provides tools for describing and computing parsimony and
weighted parsimony, and for describing allowed characters and ambiguity codes.

:class:`AmbiguityMap` stores a collection of ``Characters`` (for example the strings
``'A'``, ``'G'``, ``'C'``, and ``'T'``) that are considered valid states, as well as
a mapping between other ``Characters`` ('ambiguity codes') and subsets of these allowed
characters. This type allows forward- and backward-lookup of ambiguity codes.

:class:`TransitionModel` describes transition weights between an ordered collection of
``Characters``, and provides methods for computing weighted parsimony. Functions in
``historydag.parsimony`` accept ``TransitionModel`` objects to customize the implementation
of Sankoff, for example. This class has two subclasses: :class:`UnitTransitionModel`,
which describes unit transition costs between non-identical ``Characters``,
and :class:`SitewiseTransitionModel`, which allows transition costs to depend on the
location in a sequence in which a transition occurs.
"""
import numpy as np
import random
import historydag.utils as utils
from historydag.utils import AddFuncDict, Label
import Bio.Data.IUPACData
from frozendict import frozendict
from typing import (
    Generator,
    Tuple,
    Callable,
    Iterable,
    Sequence,
    Any,
    Mapping,
    TYPE_CHECKING,
)

Character = Any
CharacterSequence = Sequence[Character]

if TYPE_CHECKING:
    from historydag.dag import HistoryDagNode
    from historydag.compact_genome import CompactGenome


class AmbiguityMap:
    """Implements a bijection between a subset of the power set of an alphabet
    of bases, and an expanded alphabet of ambiguity codes.

    To look up the set of bases represented by an ambiguity code, use the object like a dictionary.
    To look up an ambiguity code representing a set of bases, use the ``reversed`` attribute.

    Args:
        ambiguity_character_map: A mapping from ambiguity codes to collections of bases represented by that code.
        bases: A collection of bases. If not provided, this will be inferred from the ambiguity_character_map.
        reversed_defaults: When `ambiguity_character_map` is not injective (i.e. multiple ambiguous characters
            map to the same set of bases) one might want to specify which character that set of bases maps to in
            the reversed map. If not provided, the choice will be made arbitrarily.
    """

    def __init__(
        self,
        ambiguity_character_map: Mapping[Character, Iterable[Character]],
        bases: Iterable[Character] = None,
        reversed_defaults: Mapping[Iterable[Character], Character] = {},
    ):
        ambiguous_values = {
            char: frozenset(bases) for char, bases in ambiguity_character_map.items()
        }
        if bases is None:
            self.bases = frozenset(
                base for bset in self.ambiguous_values for base in bset
            )
        else:
            self.bases = frozenset(bases)

        ambiguous_values.update({base: frozenset({base}) for base in self.bases})
        self.ambiguous_values = frozendict(ambiguous_values)
        reversed_map = {bases: char for char, bases in self.ambiguous_values.items()}
        reversed_map.update(
            {
                frozenset(key): char
                for key, char in reversed_defaults.items()
                if frozenset(key) in reversed_map
            }
        )
        self.reversed = ReversedAmbiguityMap(reversed_map)
        self.uninformative_chars = frozenset(
            char
            for char, base_set in self.ambiguous_values.items()
            if base_set == self.bases
        )

    def __getitem__(self, key):
        try:
            return self.ambiguous_values[key]
        except KeyError:
            raise KeyError(f"{key} is not a valid ambiguity code for this map.")

    def __iter__(self):
        return self.ambiguous_values.__iter__()

    def items(self):
        return self.ambiguous_values.items()

    def is_ambiguous(self, sequence: CharacterSequence) -> bool:
        """Returns whether the provided sequence contains any non-base
        characters (such as ambiguity codes)."""
        return any(code not in self.bases for code in sequence)

    def sequence_resolutions(
        self, sequence: CharacterSequence
    ) -> Generator[CharacterSequence, None, None]:
        """Iterates through possible disambiguations of sequence, recursively.

        Recursion-depth-limited by number of ambiguity codes in
        sequence, not sequence length.
        """

        def _sequence_resolutions(sequence, _accum=""):
            if sequence:
                for index, base in enumerate(sequence):
                    if base in self.bases:
                        _accum += base
                    else:
                        for newbase in self.ambiguous_values[base]:
                            yield from _sequence_resolutions(
                                sequence[index + 1 :], _accum=(_accum + newbase)
                            )
                        return
            yield _accum

        return _sequence_resolutions(sequence)

    def get_sequence_resolution_func(
        self, field_name: str
    ) -> Callable[[Label], Generator[Label, None, None]]:
        """Returns a function which takes a Label, and returns a generator on
        labels containing all possible resolutions of the sequence in that
        node's label's ``field_name`` attribute."""

        @utils.explode_label(field_name)
        def sequence_resolutions(
            sequence: CharacterSequence,
        ) -> Generator[CharacterSequence, None, None]:
            return self.sequence_resolutions(sequence)

        return sequence_resolutions

    def sequence_resolution_count(self, sequence: CharacterSequence) -> int:
        """Count the number of possible sequence resolutions.

        Equivalent to the number of items yielded by
        :meth:`sequence_resolutions`.
        """
        base_options = [
            len(self.ambiguous_values[base])
            for base in sequence
            if base in self.ambiguous_values
        ]
        return utils.prod(base_options)

    def get_sequence_resolution_count_func(
        self, field_name: str
    ) -> Callable[[Label], int]:
        """Returns a function taking a Label and returning the number of
        resolutions for the sequence held in the label's ``field_name``
        attribute."""

        @utils.access_field(field_name)
        def sequence_resolution_count(sequence) -> int:
            return self.sequence_resolution_count(sequence)

        return sequence_resolution_count


class ReversedAmbiguityMap(frozendict):
    """A subclass of frozendict for holding the reversed map in a
    :class:`AmbiguityMap` instance."""

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            raise KeyError(f"No ambiguity code is defined for the set of bases {key}")


_ambiguous_dna_values_gap_as_char = Bio.Data.IUPACData.ambiguous_dna_values.copy()
_ambiguous_dna_values_gap_as_char.update({"?": "GATC-", "-": "-"})
_ambiguous_dna_values = Bio.Data.IUPACData.ambiguous_dna_values.copy()
_ambiguous_dna_values.update({"?": "GATC", "-": "GATC"})

standard_nt_ambiguity_map = AmbiguityMap(
    _ambiguous_dna_values, "AGCT", reversed_defaults={"AGCT": "N"}
)
"""The standard IUPAC nucleotide ambiguity map, in which 'N', '?', and '-' are
all considered fully ambiguous."""

standard_nt_ambiguity_map_gap_as_char = AmbiguityMap(
    _ambiguous_dna_values_gap_as_char, "AGCT-"
)
"""The standard IUPAC nucleotide ambiguity map, in which '-' is treated as a
fifth base, '?' is fully ambiguous, and the ambiguity 'N' excludes '-'."""


class TransitionModel:
    """A class describing a collection of states, and the transition costs
    between them, for weighted parsimony.

    In addition to them methods defined below, there are also attributes which are created by
    the constructor, which should be ensured to be correct in any subclass of TransitionModel:

    * ``base_indices`` is a dictionary mapping bases in ``self.bases`` to their indices
    * ``mask_vectors`` is a dictionary mapping ambiguity codes (including non-ambiguous bases)
    to vectors of floats which are 0 at indices compatible with the ambiguity code, and infinity
    otherwise.
    * ``bases`` is a tuple recording the correct order of unambiguous bases

    Args:
        bases: An ordered collection of valid character states
        transition_weights: A matrix describing transition costs between bases, with index order (from_base, to_base)
        ambiguity_map: An :class:`AmbiguityMap` object describing ambiguity codes. If not provided, the default
            mapping depends on the provided bases. If provided bases are ``AGCT``, then
            :class:`standard_nt_ambiguity_map` is used. If bases are ``AGCT-``, then
            :class:`standard_nt_ambiguity_map_gap_as_char` is used. Otherwise, it is assumed that no ambiguity
            codes are allowed. To override these defaults, provide an :class:`AmbiguityMap` explicitly.
    """

    def __init__(
        self,
        bases: CharacterSequence = tuple("AGCT"),
        transition_weights: Sequence[Sequence[float]] = None,
        ambiguity_map: AmbiguityMap = None,
    ):
        self.bases = tuple(bases)
        self.base_indices = frozendict({base: idx for idx, base in enumerate(bases)})
        n = len(self.bases)
        if transition_weights is None:
            yey = np.ones([n, n]) - np.eye(n)
            self.transition_weights = yey
        elif len(transition_weights) != n or len(transition_weights[0]) != n:
            raise ValueError(
                "transition_weights must be a nxn matrix, with n=len(bases)"
            )
        else:
            self.transition_weights = transition_weights
        if ambiguity_map is None:
            if set(self.bases) == set("AGCT"):
                self.ambiguity_map = standard_nt_ambiguity_map
            elif set(self.bases) == set("AGCT-"):
                self.ambiguity_map = standard_nt_ambiguity_map_gap_as_char
            else:
                self.ambiguity_map = AmbiguityMap({}, self.bases)
        else:
            if not isinstance(ambiguity_map, AmbiguityMap):
                raise ValueError(
                    "ambiguity_map, if provided, must be a "
                    "historydag.parsimony.AmbiguityMap object"
                )
            if ambiguity_map.bases != set(self.bases):
                raise ValueError(
                    "ambiguity_map.bases does not match bases provided to "
                    "TransitionModel constructor"
                )
            self.ambiguity_map = ambiguity_map

        # This is applicable even when diagonal entries in transition rate matrix are
        # nonzero, since it is only a mask on allowable sites based on each base.
        self.mask_vectors = {
            code: np.array(
                [
                    0 if base in self.ambiguity_map[code] else float("inf")
                    for base in self.bases
                ]
            )
            for code in self.ambiguity_map
        }

    def get_adjacency_array(self, seq_len: int) -> np.array:
        """Return an adjacency array for a sequence of length ``seq_len``.

        This is a matrix containing transition costs, with index order
        (site, from_base, to_base)
        """
        return np.array([self.transition_weights] * seq_len)

    def get_ambiguity_from_tuple(self, tup: Tuple) -> Character:
        """Retrieve an ambiguity code encoded by tup, which encodes a set of
        bases with a tuple of 0's and 1's.

        For example, if ``self.bases`` is 'AGCT'``, then ``(0, 1, 1,
        1)`` would return `A`.
        """
        return self.ambiguity_map.reversed[
            frozenset(self.bases[i] for i, flag in enumerate(tup) if flag == 0)
        ]

    def character_distance(
        self, parent_char: Character, child_char: Character, site: int = None
    ) -> float:
        """Return the transition cost from parent_char to child_char, two
        unambiguous characters.

        keyword argument ``site`` is ignored in this base class.
        """
        return self.transition_weights[self.base_indices[parent_char]][
            self.base_indices[child_char]
        ]

    def weighted_hamming_distance(
        self, parent_seq: CharacterSequence, child_seq: CharacterSequence
    ) -> float:
        """Return the sum of sitewise transition costs, from parent_seq to
        child_seq.

        Both parent_seq and child_seq are expected to be unambiguous.
        """
        if len(parent_seq) != len(child_seq):
            raise ValueError("sequence lengths do not match!")
        return sum(
            self.character_distance(pchar, cchar, site=(idx + 1))
            for idx, (pchar, cchar) in enumerate(zip(parent_seq, child_seq))
        )

    def weighted_cg_hamming_distance(
        self, parent_cg: "CompactGenome", child_cg: "CompactGenome"
    ) -> float:
        """Return the sum of sitewise transition costs, from parent_cg to
        child_cg.

        Both parent_seq and child_seq are expected to be unambiguous,
        but sites where parent_cg and child_cg both match their
        reference sequence are ignored, so this method is not suitable
        for weighted parsimony for transition matrices that contain
        nonzero entries along the diagonal.
        """
        if parent_cg.reference != child_cg.reference:
            raise ValueError("Reference sequences do not match!")
        s1 = set(parent_cg.mutations.keys())
        s2 = set(child_cg.mutations.keys())
        return sum(
            self.character_distance(
                parent_cg.get_site(site), child_cg.get_site(site), site=site
            )
            for site in s1 | s2
        )

    def min_character_mutation(
        self,
        parent_char: Character,
        child_char: Character,
        site=None,
        **kwargs,
    ) -> float:
        """Allowing child_char to be ambiguous, returns the minimum possible
        transition weight between parent_char and child_char, preceded in a
        tuple by the unambiguous child character that achieves that weight.

        Keyword argument ``site`` is ignored in this base class.
        """
        child_set = sorted(list(self.ambiguity_map[child_char]))
        p_idx = self.base_indices[parent_char]
        return min(
            (
                (cbase, self.transition_weights[p_idx][self.base_indices[cbase]])
                for cbase in child_set
            ),
            key=lambda it: it[-1],
        )

    def min_character_distance(
        self, parent_char: Character, child_char: Character, site=None
    ) -> float:
        """Allowing child_char to be ambiguous, returns the minimum possible
        transition weight between parent_char and child_char.

        Keyword argument ``site`` is ignored in this base class.
        """
        return self.min_character_mutation(
            parent_char,
            child_char,
            site=site,
        )[1]

    def min_weighted_hamming_distance(
        self, parent_seq: CharacterSequence, child_seq: CharacterSequence
    ) -> float:
        """Assuming the child_seq may contain ambiguous characters, returns the
        minimum possible transition weight between parent_seq and child_seq."""
        if len(parent_seq) != len(child_seq):
            raise ValueError("sequence lengths do not match!")
        return sum(
            self.min_character_distance(pchar, cchar, site=(idx + 1))
            for idx, (pchar, cchar) in enumerate(zip(parent_seq, child_seq))
        )

    def min_weighted_cg_hamming_distance(
        self, parent_cg: "CompactGenome", child_cg: "CompactGenome"
    ) -> float:
        """Return the sum of sitewise transition costs, from parent_cg to
        child_cg.

        child_cg may contain ambiguous characters, and in this case
        those sites will contribute the minimum possible transition cost
        to the returned value.

        Sites where parent_cg and child_cg both match their reference
        sequence are ignored, so this method is not suitable for
        weighted parsimony for transition matrices that contain nonzero
        entries along the diagonal.
        """
        if parent_cg.reference != child_cg.reference:
            raise ValueError("Reference sequences do not match!")
        s1 = set(parent_cg.mutations.keys())
        s2 = set(child_cg.mutations.keys())
        return sum(
            self.min_character_distance(
                parent_cg.get_site(site), child_cg.get_site(site), site=site
            )
            for site in s1 | s2
        )

    def weighted_hamming_edge_weight(
        self, field_name: str
    ) -> Callable[["HistoryDagNode", "HistoryDagNode"], float]:
        """Returns a function for computing weighted hamming distance between
        two nodes' sequences.

        Args:
            field_name: The name of the node label field which contains sequences.

        Returns:
            A function accepting two :class:`HistoryDagNode` objects ``n1`` and ``n2`` and returning
            a float: the transition cost from ``n1.label.<field_name>`` to ``n2.label.<field_name>``, or 0 if
            n1 is the UA node.
        """

        @utils.access_nodefield_default(field_name, 0)
        def edge_weight(parent, child):
            return self.weighted_hamming_distance(parent, child)

        return edge_weight

    def weighted_cg_hamming_edge_weight(
        self,
        field_name: str,
        count_root_muts: bool = True,
    ) -> Callable[["HistoryDagNode", "HistoryDagNode"], float]:
        """Returns a function for computing weighted hamming distance between
        two nodes' compact genomes.

        Args:
            field_name: The name of the node label field which contains compact genomes.
            count_root_muts: If True, the reference sequence of root nodes' compact
                genomes is considered to be the ancestral sequence of each history,
                and mutations from the reference to the history root CG are counted.

        Returns:
            A function accepting two :class:`HistoryDagNode` objects ``n1`` and ``n2`` and returning
            a float: the transition cost from ``n1.label.<field_name>`` to ``n2.label.<field_name>``, or
            the number of mutations between the reference and the CG of n2 if n1 is the UA node.

        Sites where parent_cg and child_cg
        both match their reference sequence are ignored, so this method is not suitable for weighted parsimony
        for transition matrices that contain nonzero entries along the diagonal.
        """

        def edge_weight(parent, child):
            # Get the number of mutations on a branch pointing to the UA node
            if parent.is_ua_node():
                if count_root_muts:
                    return len(getattr(child.label, field_name).mutations)
                else:
                    return 0
            else:
                return self.weighted_cg_hamming_distance(
                    getattr(parent.label, field_name), getattr(child.label, field_name)
                )

        return edge_weight

    def min_weighted_hamming_edge_weight(
        self, field_name: str
    ) -> Callable[["HistoryDagNode", "HistoryDagNode"], float]:
        """Returns a function for computing weighted hamming distance between
        two nodes' sequences.

        If the child node is a leaf node, and its sequence contains ambiguities, the minimum possible
        transition cost from parent to child will be returned.

        Args:
            field_name: The name of the node label field which contains sequences.

        Returns:
            A function accepting two :class:`HistoryDagNode` objects ``n1`` and ``n2`` and returning
            a float: the transition cost from ``n1.label.<field_name>`` to ``n2.label.<field_name>``, or 0 if
            n1 is the UA node.
        """

        def edge_weight(parent, child):
            if parent.is_ua_node():
                return 0
            elif child.is_leaf():
                return self.min_weighted_hamming_distance(
                    getattr(parent.label, field_name), getattr(child.label, field_name)
                )
            else:
                return self.weighted_hamming_distance(
                    getattr(parent.label, field_name), getattr(child.label, field_name)
                )

        return edge_weight

    def min_weighted_cg_hamming_edge_weight(
        self,
        field_name: str,
        count_root_muts: bool = True,
    ) -> Callable[["HistoryDagNode", "HistoryDagNode"], float]:
        """Returns a function for computing weighted hamming distance between
        two nodes' compact genomes.

        If the child node is a leaf node, and its compact genome contains ambiguities, the minimum possible
        transition cost from parent to child will be returned by this function.

        Args:
            field_name: The name of the node label field which contains compact genomes.
            count_root_muts: If True, the reference sequence of root nodes' compact
                genomes is considered to be the ancestral sequence of each history,
                and mutations from the reference to the history root CG are counted.

        Returns:
            A function accepting two :class:`HistoryDagNode` objects ``n1`` and ``n2`` and returning
            a float: the transition cost from ``n1.label.<field_name>`` to ``n2.label.<field_name>``, or 0 if
            n1 is the UA node.
        """
        unambiguous_edge_weight = self.weighted_cg_hamming_edge_weight(
            field_name,
            count_root_muts=count_root_muts,
        )

        def edge_weight(parent, child):
            if child.is_leaf():
                return self.min_weighted_cg_hamming_distance(
                    getattr(parent.label, field_name), getattr(child.label, field_name)
                )
            else:
                return unambiguous_edge_weight(parent, child)

        return edge_weight

    def get_weighted_cg_parsimony_countfuncs(
        self,
        field_name: str,
        leaf_ambiguities: bool = False,
        count_root_muts: bool = True,
        name: str = "WeightedParsimony",
    ) -> AddFuncDict:
        """Create a :class:`historydag.utils.AddFuncDict` object for counting
        weighted parsimony in a HistoryDag with labels containing compact
        genomes.

        Args:
            field_name: the label field name in which compact genomes can be found
            leaf_ambiguities: if True, leaf compact genomes are permitted to contain ambiguity codes
            count_root_muts: If True, the reference sequence of root nodes' compact
                genomes is considered to be the ancestral sequence of each history,
                and mutations from the reference to the history root CG are counted.
            name: the name for the returned AddFuncDict object
        """
        if leaf_ambiguities:
            edge_weight = self.min_weighted_cg_hamming_edge_weight(
                field_name, count_root_muts=count_root_muts
            )
        else:
            edge_weight = self.weighted_cg_hamming_edge_weight(
                field_name, count_root_muts=count_root_muts
            )
        return AddFuncDict(
            {
                "start_func": lambda n: 0,
                "edge_weight_func": edge_weight,
                "accum_func": sum,
            },
            name=name,
        )

    def get_weighted_parsimony_countfuncs(
        self,
        field_name: str,
        leaf_ambiguities: bool = False,
        name: str = "WeightedParsimony",
    ):
        """Create a :class:`historydag.utils.AddFuncDict` object for counting
        weighted parsimony in a HistoryDag with labels containing sequences.

        Args:
            field_name: the label field name in which sequences can be found
            leaf_ambiguities: if True, leaf sequences are permitted to contain ambiguity codes
            name: the name for the returned AddFuncDict object
        """
        if leaf_ambiguities:
            edge_weight = self.min_weighted_hamming_edge_weight(field_name)
        else:
            edge_weight = self.weighted_hamming_edge_weight(field_name)
        return AddFuncDict(
            {
                "start_func": lambda n: 0,
                "edge_weight_func": edge_weight,
                "accum_func": sum,
            },
            name=name,
        )


class UnitTransitionModel(TransitionModel):
    """A subclass of :class:`TransitionModel` to be used when all non-identical
    transitions have unit cost.

    Args:
        bases: An ordered container of allowed character states
        ambiguity_map: A :class:`AmbiguityMap` object describing allowed ambiguity codes.
            If not provided, a default will be chosen by the same method as the constructor for
            :class:`TransitionModel`.
    """

    def __init__(
        self, bases: CharacterSequence = "AGCT", ambiguity_map: AmbiguityMap = None
    ):
        super().__init__(bases=bases, ambiguity_map=None)

    def character_distance(
        self, parent_char: Character, child_char: Character, site: int = None
    ) -> int:
        """Return the transition cost from parent_char to child_char, two
        unambiguous characters.

        keyword argument ``site`` is ignored in this subclass.
        """
        return int(parent_char != child_char)

    def min_character_mutation(
        self,
        parent_char: Character,
        child_char: Character,
        site=None,
        randomize=False,
        **kwargs,
    ) -> float:
        """Allowing child_char to be ambiguous, returns the minimum possible
        transition weight between parent_char and child_char, preceded in a
        tuple by the unambiguous child character that achieves that weight.

        Keyword argument ``site`` is ignored in this base class.
        """
        if parent_char == child_char:
            return (child_char, 0)
        else:
            options = self.ambiguity_map[child_char]
            if parent_char in options:
                return (parent_char, 0)
            else:
                if randomize:
                    base = random.choice(list(sorted(options)))
                else:
                    # arbitrary, but deterministic
                    base = min(options)
                return (base, 1)

    def get_weighted_cg_parsimony_countfuncs(
        self,
        field_name: str,
        leaf_ambiguities: AmbiguityMap = False,
        name: str = "HammingParsimony",
    ) -> AddFuncDict:
        """Create a :class:`historydag.utils.AddFuncDict` object for counting
        parsimony in a HistoryDag with labels containing compact genomes.

        Args:
            field_name: the label field name in which compact genomes can be found
            leaf_ambiguities: if True, leaf compact genomes are permitted to contain ambiguity codes
            name: the name for the returned AddFuncDict object
        """
        return super().get_weighted_cg_parsimony_countfuncs(
            field_name, leaf_ambiguities=leaf_ambiguities, name=name
        )

    def get_weighted_parsimony_countfuncs(
        self,
        field_name: str,
        leaf_ambiguities: AmbiguityMap = False,
        name: str = "HammingParsimony",
    ) -> AddFuncDict:
        """Create a :class:`historydag.utils.AddFuncDict` object for counting
        weighted parsimony in a HistoryDag with labels containing sequences.

        Args:
            field_name: the label field name in which sequences can be found
            leaf_ambiguities: if True, leaf sequences are permitted to contain ambiguity codes
            name: the name for the returned AddFuncDict object
        """
        return super().get_weighted_parsimony_countfuncs(
            field_name, leaf_ambiguities=leaf_ambiguities, name=name
        )


class SitewiseTransitionModel(TransitionModel):
    """A subclass of :class:`TransitionModel` to be used when transition costs
    depend on site.

    Args:
        bases: An ordered container of allowed character states
        transition_matrix: A matrix of transition costs between provided bases, in which
            index order is (site, from_base, to_base)
        ambiguity_map: An :class:`AmbiguityMap` object describing ambiguous characters. If not provided
            a default will be chosen as in the constructor for :class:`TransitionModel`.
    """

    def __init__(
        self,
        bases: CharacterSequence = "AGCT",
        transition_matrix: Sequence[Sequence[Sequence[float]]] = None,
        ambiguity_map: AmbiguityMap = None,
    ):
        assert transition_matrix.shape[1:] == (len(bases), len(bases))
        super().__init__(bases=bases, ambiguity_map=None)
        self.transition_weights = None
        self.sitewise_transition_matrix = transition_matrix
        self._seq_len = len(self.sitewise_transition_matrix)

    def character_distance(
        self, parent_char: Character, child_char: Character, site: int
    ) -> float:
        """Returns the transition cost from ``parent_char`` to ``child_char``,
        assuming each character is at the one-based ``site`` in their
        respective sequences.

        Both parent_char and child_char are expected to be unambiguous.
        """
        return self.sitewise_transition_matrix[site - 1][parent_char][child_char]

    def min_character_mutation(
        self,
        parent_char: Character,
        child_char: Character,
        site=None,
        **kwargs,
    ) -> float:
        """Allowing child_char to be ambiguous, returns the minimum possible
        transition weight between parent_char and child_char, preceded in a
        tuple by the unambiguous child character that achieves that weight.

        Keyword argument ``site`` is ignored in this base class.
        """
        child_set = sorted(list(self.ambiguity_map[child_char]))
        p_idx = self.base_indices[parent_char]
        return min(
            (
                (
                    cbase,
                    self.transition_weights[site - 1][p_idx][self.base_indices[cbase]],
                )
                for cbase in child_set
            ),
            key=lambda it: it[-1],
        )

    def get_adjacency_array(self, seq_len):
        """Return an adjacency array for a sequence of length ``seq_len``.

        This is a matrix containing transition costs, with index order
        (site, from_base, to_base)
        """
        if seq_len != self._seq_len:
            raise ValueError(
                f"SitewiseTransitionModel instance supports sequence length of {self._seq_len}"
            )
        return self.sitewise_transition_matrix


default_nt_transitions = UnitTransitionModel(bases="AGCT")
"""A transition model for nucleotides using unit transition weights, and the
standard IUPAC ambiguity codes, with base order ``AGCT`` and ``N``, ``?`` and
``-`` treated as fully ambiguous characters."""

default_nt_gaps_transitions = UnitTransitionModel(bases="AGCT-")
"""A transition model for nucleotides using unit transition weights, and the
standard IUPAC nucleotide ambiguity map, with '-' is treated as a fifth base,
'?' fully ambiguous, and the ambiguity 'N' excluding only '-'."""

hamming_edge_weight = default_nt_transitions.weighted_hamming_edge_weight("sequence")
"""An edge weight function accepting (parent, child) pairs of HistoryDagNodes,
and returning the hamming distance between the sequences stored in the
``sequence`` attributes of their labels."""
hamming_edge_weight_ambiguous_leaves = (
    default_nt_transitions.min_weighted_hamming_edge_weight("sequence")
)
"""An edge weight function accepting (parent, child) pairs of HistoryDagNodes,
and returning the hamming distance between the sequences stored in the
``sequence`` attributes of their labels.

This edge weight function allows leaf nodes to have ambiguous sequences.
"""

hamming_cg_edge_weight = default_nt_transitions.weighted_cg_hamming_edge_weight(
    "compact_genome"
)
"""An edge weight function accepting (parent, child) pairs of HistoryDagNodes,
and returning the hamming distance between the compact genomes stored in the
``compact_genome`` attributes of their labels."""

hamming_cg_edge_weight_ambiguous_leaves = (
    default_nt_transitions.min_weighted_cg_hamming_edge_weight("compact_genome")
)
"""An edge weight function accepting (parent, child) pairs of HistoryDagNodes,
and returning the hamming distance between the compact genomes stored in the
``compact_genome`` attributes of their labels.

This edge weight function allows leaf nodes to have ambiguous compact
genomes.
"""

hamming_distance_countfuncs = default_nt_transitions.get_weighted_parsimony_countfuncs(
    "sequence", leaf_ambiguities=False
)
"""Provides functions to count hamming distance parsimony when leaf sequences
may be ambiguous.

For use with :meth:`historydag.AmbiguousLeafSequenceHistoryDag.weight_count`.
"""

leaf_ambiguous_hamming_distance_countfuncs = (
    default_nt_transitions.get_weighted_parsimony_countfuncs(
        "sequence", leaf_ambiguities=True
    )
)
"""Provides functions to count hamming distance parsimony when leaf sequences
may be ambiguous.

For use with :meth:`historydag.AmbiguousLeafSequenceHistoryDag.weight_count`.
"""


compact_genome_hamming_distance_countfuncs = (
    default_nt_transitions.get_weighted_cg_parsimony_countfuncs(
        "compact_genome", leaf_ambiguities=False
    )
)
"""Provides functions to count hamming distance parsimony when sequences are
stored as CompactGenomes.

For use with :meth:`historydag.CGHistoryDag.weight_count`.
"""


leaf_ambiguous_compact_genome_hamming_distance_countfuncs = (
    default_nt_transitions.get_weighted_cg_parsimony_countfuncs(
        "compact_genome", leaf_ambiguities=True
    )
)
"""Provides functions to count hamming distance parsimony when leaf compact
genomes may be ambiguous.

For use with :meth:`historydag.AmbiguousLeafCGHistoryDag.weight_count`.
"""



from historydag import utils
from historydag import dag as hdag
import ete3
from Bio.Data.IUPACData import ambiguous_dna_values
from historydag import parsimony_utils

bases = "AGCT-"
ambiguous_dna_values.update({"?": "GATC-", "-": "-"})


def sequence_resolutions(sequence):
    """Iterates through possible disambiguations of sequence, recursively.
    Recursion-depth-limited by number of ambiguity codes in
    sequence, not sequence length.
    """

    def _sequence_resolutions(sequence, _accum=""):
        if sequence:
            for index, base in enumerate(sequence):
                if base in bases:
                    _accum += base
                else:
                    for newbase in ambiguous_dna_values[base]:
                        yield from _sequence_resolutions(
                            sequence[index + 1 :], _accum=(_accum + newbase)
                        )
                    return
        yield _accum

    return _sequence_resolutions(sequence)


def disambiguate_sitewise(tree: ete3.TreeNode) -> ete3.TreeNode:
    """Resolve tree and return list of all possible resolutions"""
    code_vectors = {
        code: [
            0 if base in ambiguous_dna_values[code] else float("inf") for base in bases
        ]
        for code in ambiguous_dna_values
    }
    cost_adjust = {
        base: [int(not i == j) for j in range(5)] for i, base in enumerate(bases)
    }

    for node in tree.traverse(strategy="postorder"):

        def cvup(node, site):
            cv = code_vectors[node.sequence[site]].copy()
            if not node.is_leaf():
                for i in range(5):
                    for child in node.children:
                        cv[i] += min(
                            [
                                sum(v)
                                for v in zip(child.cvd[site], cost_adjust[bases[i]])
                            ]
                        )
            return cv

        # Make dictionary of cost vectors for each site
        node.cvd = {site: cvup(node, site) for site in range(len(node.sequence))}

    disambiguated = [tree.copy()]
    ambiguous = True
    while ambiguous:
        ambiguous = False
        treesindex = 0
        while treesindex < len(disambiguated):
            tree2 = disambiguated[treesindex]
            treesindex += 1
            for node in tree2.traverse(strategy="preorder"):
                ambiguous_sites = [
                    site for site, code in enumerate(node.sequence) if code not in bases
                ]
                if not ambiguous_sites:
                    continue
                else:
                    ambiguous = True
                    # Adjust cost vectors for ambiguous sites base on above
                    if not node.is_root():
                        for site in ambiguous_sites:
                            base_above = node.up.sequence[site]
                            node.cvd[site] = [
                                sum(v)
                                for v in zip(node.cvd[site], cost_adjust[base_above])
                            ]
                    option_dict = {site: "" for site in ambiguous_sites}
                    # Enumerate min-cost choices
                    for site in ambiguous_sites:
                        min_cost = min(node.cvd[site])
                        min_cost_sites = [
                            bases[i]
                            for i, val in enumerate(node.cvd[site])
                            if val == min_cost
                        ]
                        option_dict[site] = "".join(min_cost_sites)

                    sequences = list(utils._options(option_dict, node.sequence))
                    # Give this tree the first sequence, append copies with all
                    # others to disambiguated.
                    numseqs = len(sequences)
                    for idx, sequence in enumerate(sequences):
                        node.sequence = sequence
                        if idx < numseqs - 1:
                            disambiguated.append(tree2.copy())
                    break
    for tree in disambiguated:
        for node in tree.traverse():
            try:
                node.del_feature("cvd")
            except KeyError:
                pass
    return disambiguated


def disambiguate(
    tree: ete3.Tree,
    dist_func=parsimony_utils.default_nt_transitions.weighted_hamming_distance,
):
    """Resolve ambiguous bases using a two-pass Sankoff Algorithm on entire tree and entire sequence at each node.
    This does not disambiguate sitewise, so trees with many ambiguities may make this run very slowly.
    Returns a list of all possible disambiguations, minimizing the passed distance function dist_func.
    """
    # First pass of Sankoff: compute cost vectors
    for node in tree.traverse(strategy="postorder"):
        node.add_feature(
            "costs", [[seq, 0] for seq in sequence_resolutions(node.sequence)]
        )
        if not node.is_leaf():
            for seq_cost in node.costs:
                for child in node.children:
                    seq_cost[1] += min(
                        [
                            dist_func(seq_cost[0], child_seq) + child_cost
                            for child_seq, child_cost in child.costs
                        ]
                    )
    disambiguate_queue = [tree]
    disambiguated = []

    def incremental_disambiguate(ambig_tree):
        for node in ambig_tree.traverse(strategy="preorder"):
            if not parsimony_utils.default_nt_transitions.ambiguity_map.is_ambiguous(
                node.sequence
            ):
                continue
            else:
                if not node.is_root():
                    node.costs = [
                        [
                            sequence,
                            cost + dist_func(node.up.sequence, sequence),
                        ]
                        for sequence, cost in node.costs
                    ]
                min_cost = min([cost for _, cost in node.costs])
                for resolved_sequence in [
                    sequence for sequence, cost in node.costs if cost == min_cost
                ]:
                    node.sequence = resolved_sequence
                    disambiguate_queue.append(ambig_tree.copy())
                return
        disambiguated.append(ambig_tree)

    while disambiguate_queue:
        incremental_disambiguate(disambiguate_queue.pop())
    for tree in disambiguated:
        for node in tree.traverse():
            try:
                node.del_feature("costs")
            except (AttributeError, KeyError):
                pass
    return disambiguated


newick_tree2 = (
    "((12[&&NHX:name=12:sequence=TT],"
    "(6[&&NHX:name=6:sequence=CG],"
    "7[&&NHX:name=7:sequence=AC])5[&&NHX:name=5:sequence=??])"
    "3[&&NHX:name=3:sequence=?T],8[&&NHX:name=8:sequence=AA],"
    "(11[&&NHX:name=11:sequence=AG],10[&&NHX:name=10:sequence=GT])"
    "9[&&NHX:name=9:sequence=?T])2[&&NHX:name=2:sequence=?T];"
)
tree2 = ete3.TreeNode(newick=newick_tree2, format=1)
dags = [
    hdag.history_dag_from_newicks(newicklist, ["sequence"])
    for newicklist in [[newick_tree2]]
]


def treeprint(tree: ete3.TreeNode):
    tree = tree.copy()
    for node in tree.traverse():
        node.name = node.sequence
    return tree.write(format=8)


def test_expand_ambiguities():
    for dag in dags:
        cdag = dag.copy()
        print(cdag.count_histories())
        cdag.explode_nodes()
        print(cdag.count_histories())
        cdag.trim_optimal_weight()
        cdag._check_valid()
        print(cdag.count_histories())
        print(cdag.weight_count())
        checkset = {
            hdag.from_tree(tree, ["sequence"]).to_newick()
            for cladetree in dag.get_histories()
            for tree in disambiguate(cladetree.to_ete(features=["sequence"]))
        }
        print(len(checkset))
        assert checkset == {cladetree.to_newick() for cladetree in cdag.get_histories()}


#     newickset = {treeprint(tree) for tree in utils.disambiguate(tree2)}
#     correctset = {
#         "((((T)C,(C,A)C)C,A,(A,G)G)G);",
#         "((((T)C,(C,A)A)A,A,(A,G)A)A);",
#         "((((T)C,(C,A)C)C,A,(A,G)A)A);",
#     }
#     if not newickset == correctset:
#         missing = correctset - newickset
#         wrong = newickset - correctset
#         print(
#             f"\nDisambiguate function is missing {missing}\n"
#             f"and came up with these incorrect trees: {wrong}"
#         )
#         raise ValueError("Invalid Disambiguation")



"""A module implementing Sankoff Algorithm."""
import random
import ete3
import numpy as np
from itertools import product
import historydag.utils as utils
from historydag.utils import load_fasta
from historydag.dag import (
    history_dag_from_histories,
    history_dag_from_etes,
    HistoryDag,
)
from copy import deepcopy
import historydag.parsimony_utils as parsimony_utils


def replace_label_attr(original_label, list_of_replacements={}):
    """Generalizes :meth: ``_replace()`` for namedtuple datatype to replace
    multiple fields at once, and by string rather than as a keyword argument.

    Caveat: the keys of ``list_of_replacements`` dict should be existing fields in the namedtuple object for ``original_label``
    """
    fields = original_label._asdict()
    fields.update(list_of_replacements)
    return type(original_label)(**fields)


def make_weighted_hamming_count_funcs(
    transition_model=parsimony_utils.default_nt_transitions,
    allow_ambiguous_leaves=False,
    sequence_label="sequence",
):
    """Returns an :class:`AddFuncDict` for computing weighted parsimony.

    The returned ``AddFuncDict`` is for use with :class:`HistoryDag` objects whose nodes
    have unambiguous sequences stored in label attributes named ``sequence``.

    Args:

    Returns:
        :class:`AddFuncDict` object for computing weighted parsimony.
    """

    if allow_ambiguous_leaves:
        weight_func = transition_model.min_weighted_hamming_edge_weight(sequence_label)
    else:
        weight_func = transition_model.weighted_hamming_edge_weight(sequence_label)
    return utils.AddFuncDict(
        {
            "start_func": lambda n: 0,
            "edge_weight_func": weight_func,
            "accum_func": sum,
        },
        name="WeightedParsimony",
    )


def sankoff_postorder_iter_accum(
    postorder_iter, node_clade_function, child_node_function
):
    """This is a re-take of :meth:`postorder_history_accum` that is altered so
    that it does not require a complete DAG, and simplified for the specific
    Sankoff algorithm.

    Args:
        postorder_iter: iterable of nodes to traverse.
        node_clade_function: function to combine results for a given clade.
        child_node_function: function that is applied to children for a given node.
    """
    if any(postorder_iter):
        for node in postorder_iter:
            if node.is_leaf():
                node._dp_data = {
                    "cost_vectors": child_node_function(node),
                    "subtree_cost": 0,
                }
            else:
                node._dp_data = node_clade_function(
                    [
                        [
                            child_node_function(target)
                            for target in node.children(clade=clade)
                        ]
                        for clade in node.clades
                    ]
                )
        return node._dp_data
    return {"cost_vectors": [], "subtree_cost": 0}


def sankoff_upward(
    node_list,
    seq_len,
    sequence_attr_name="sequence",
    transition_model=parsimony_utils.default_nt_transitions,
    use_internal_node_sequences=False,
):
    """Compute Sankoff cost vectors at nodes in a postorder traversal, and
    return best possible parsimony score of the tree.

    Args:
        node_list: An iterator for the set of nodes to compute Sankoff on. Can be of type ``ete3.TreeNode``
            or of type ``HistoryDag``, or else a list of nodes that respects postorder traversal. This last
            option is useful when using the Sankoff algorithm to compute node labels for a subset of the
            internal nodes of a history DAG, as only the nodes in the list will be relabelled. However it
            does assume that the remaining internal nodes are in fact equipped with a sequence.
        seq_len: The length of sequence for each leaf node.
        sequence_attr_name: the field name of the label attribute that is used to store sequences on nodes.
            Default value is set to 'sequence'
        transition_model: The type of ``TransitionModel`` to use for Sankoff. See docstring for more information.
        use_internal_node_sequences: (used when tree is of type ``ete3.TreeNode``) If True, then compute the
            transition cost for sequences assigned to internal nodes.
    """

    adj_arr = transition_model.get_adjacency_array(seq_len)
    if isinstance(node_list, ete3.TreeNode):
        # First pass of Sankoff: compute cost vectors
        for node in node_list.traverse(strategy="postorder"):
            node.add_feature(
                "cost_vector",
                np.array(
                    [
                        transition_model.mask_vectors[base].copy()
                        for base in getattr(node, sequence_attr_name)
                    ]
                ),
            )
            if not node.is_leaf():
                child_costs = []
                for child in node.children:
                    stacked_child_cv = np.stack(
                        (child.cost_vector,) * len(transition_model.bases), axis=1
                    )
                    total_cost = adj_arr + stacked_child_cv
                    child_costs.append(np.min(total_cost, axis=2))
                child_cost = np.sum(child_costs, axis=0)
                node.cost_vector = child_cost
            if use_internal_node_sequences:
                node.cost_vector += np.array(
                    [
                        transition_model.mask_vectors[base]
                        for base in getattr(node, sequence_attr_name)
                    ]
                )
        return np.sum(np.min(node_list.cost_vector, axis=1))

    elif isinstance(node_list, HistoryDag):
        node_list = list(node_list.postorder())
    if isinstance(node_list, list):
        sequence_attr_idx = node_list[0].label._fields.index(sequence_attr_name)
        max_transition_cost = np.amax(adj_arr) * seq_len

        def children_cost(child_cost_vectors):
            costs = []
            for c in child_cost_vectors:
                cost = adj_arr + np.stack((c,) * len(transition_model.bases), axis=1)
                costs.append(np.min(cost, axis=2))
            return np.sum(costs, axis=0)

        def cost_vector(node):
            if node.is_leaf():
                return [
                    np.array(
                        [
                            transition_model.mask_vectors[base].copy()
                            for base in node.label[sequence_attr_idx]
                        ]
                    )
                ]
            elif isinstance(node._dp_data, dict):
                return node._dp_data["cost_vectors"]
            else:
                return [
                    np.array(
                        [
                            transition_model.mask_vectors[base].copy()
                            for base in node.label[sequence_attr_idx]
                        ]
                    )
                ]

        def accum_between_clade(list_of_clade_cvs):
            cost_vectors = []
            min_cost = float("inf")
            for choice in product(*list_of_clade_cvs):
                for cost_vector_combination in product(*[c for c in choice]):
                    cv = children_cost(cost_vector_combination)
                    cost = np.sum(np.min(cv, axis=1))
                    if (cost + max_transition_cost) < min_cost:
                        min_cost = cost
                        cost_vectors = [cv]
                    elif cost <= (min_cost + max_transition_cost) and not any(
                        [np.array_equal(cv, other_cv) for other_cv in cost_vectors]
                    ):
                        min_cost = min(cost, min_cost)
                        cost_vectors.append(cv)
            return {"cost_vectors": cost_vectors, "subtree_cost": min_cost}

        compute_val = sankoff_postorder_iter_accum(
            node_list, accum_between_clade, cost_vector
        )
        return compute_val["subtree_cost"]
    else:
        return 0


def sankoff_downward(
    dag,
    skip_ua_children=False,
    partial_node_list=None,
    sequence_attr_name="sequence",
    compute_cvs=True,
    transition_model=parsimony_utils.default_nt_transitions,
    trim=True,
):
    """Assign sequences to internal nodes of dag using a weighted Sankoff
    algorithm by exploding all possible labelings associated to each internal
    node based on its subtrees. This function alters the node labels and the
    topology of the DAG.

    Args:
        skip_ua_children: If True, all children of the UA node will remain unchanged.
        partial_node_list: If None, all internal node sequences will be re-computed. If a list of nodes
            is provided, only sequences on nodes from that list will be updated. This list must be in
            post-order.
        sequence_attr_name: The field name of the label attribute that is used to store sequences on nodes.
            Default value is set to 'sequence'
        compute_cvs: If true, compute upward sankoff cost vectors. If ``sankoff_upward`` was
            already run on the tree/dag, this may be skipped.
        transition_model: The type of ``TransitionModel`` to use for Sankoff. See docstring for more information.
        trim: If False, the history DAG will not be trimmed to express only maximally parsimonious
            histories after Sankoff.
    """

    sequence_attr_idx = next(dag.get_leaves()).label._fields.index(sequence_attr_name)
    seq_instance = next(dag.get_leaves()).label[sequence_attr_idx]
    seq_len = len(seq_instance)

    # `join_bases` is the method for creating a new instance of sequence given an iterable of bases 'n'
    if isinstance(seq_instance, str):

        def join_bases(n):
            return "".join(n)

    else:

        def join_bases(n):
            return type(seq_instance)([x for x in n])

    if partial_node_list is None:
        partial_node_list = list(dag.postorder())
    if skip_ua_children:
        ua_children = set(dag.dagroot.children())
        partial_node_list = [n for n in partial_node_list if n not in ua_children]
    if compute_cvs:
        sankoff_upward(
            partial_node_list,
            seq_len,
            sequence_attr_name=sequence_attr_name,
            transition_model=transition_model,
        )
    # save the field names/types of the label datatype for this dag
    adj_arr = transition_model.get_adjacency_array(seq_len)

    def transition_cost(seq):
        return adj_arr[
            np.arange(len(seq)), [transition_model.base_indices[s] for s in seq]
        ]

    def compute_sequence_data(cost_vector):
        """Compute all possible sequences that minimize transition costs as
        given by cost_vector.

        Returns: a list of tuples, each tuple containing:
                  - a sequence
                  - an adjacency array of costs for that sequence, and
                  - the minimum cost associated to the cost vector
        """
        all_base_indices = [[]]
        min_cost = sum(np.min(cost_vector, axis=1))
        for idx in range(seq_len):
            min_cost_indices = np.where(cost_vector[idx] == cost_vector[idx].min())[0]
            all_base_indices = [
                base_idx + [i]
                for base_idx in all_base_indices
                for i in min_cost_indices
            ]
        adj_vec = [
            adj_arr[np.arange(seq_len), base_indices]
            for base_indices in all_base_indices
        ]
        new_sequence = [
            join_bases(
                [transition_model.bases[base_index] for base_index in base_indices]
            )
            for base_indices in all_base_indices
        ]
        return list(zip(new_sequence, adj_vec, [min_cost] * len(new_sequence)))

    dag_nodes = {}
    # downward pass of Sankoff: find and assign sequence labels to each internal node
    for node in reversed(partial_node_list):
        if not (node.is_leaf() or node.is_ua_node()):
            node_data = {k: v for k, v in node._dp_data.items()}
            node_copies = {}
            node_children = set(node.children())
            node_parents = set(node.parents)
            for p in node_parents:
                if p.is_ua_node():
                    new_seq_data = [
                        y
                        for cv in node._dp_data["cost_vectors"]
                        for y in compute_sequence_data(cv)
                    ]
                else:
                    new_seq_data = [
                        y
                        for cv in node._dp_data["cost_vectors"]
                        for y in compute_sequence_data(
                            cv + transition_cost(p.label[sequence_attr_idx])
                        )
                    ]

                # only keep those node/parent/cost_vector choices that
                # achieve a minimal cost for the node/parent choice
                min_val = new_seq_data[0][-1]
                if len(new_seq_data) > 1:
                    min_val = min(*list(zip(*new_seq_data))[-1])
                for nsd in new_seq_data:
                    if (nsd[-1] <= min_val) and (nsd[0] not in node_copies):
                        new_node = node.empty_copy()
                        new_node.label = replace_label_attr(
                            node.label, {sequence_attr_name: nsd[0]}
                        )
                        new_node._dp_data = deepcopy(node_data)
                        node_copies[nsd[0]] = new_node
            for c in node_children:
                c.parents.remove(node)
            for p in node_parents:
                p.remove_edge_by_clade_and_id(node, node.clade_union())
            # add all new copies of current node(with alt sequence labels) into the dag
            for new_sequence, new_node in node_copies.items():
                if new_node in dag_nodes:
                    new_node = dag_nodes[new_node]
                for c in node_children:
                    new_node.add_edge(c)
                    c.parents.add(new_node)
                for parent in node_parents:
                    parent.add_edge(new_node)
                new_node.parents.update(node_parents)
                dag_nodes[new_node] = new_node

    dag.recompute_parents()
    # still need to trim the dag since the final addition of all
    # parents/children to new nodes can yield suboptimal choices

    weight_func = utils.access_nodefield_default(sequence_attr_name, 0)(
        transition_model.weighted_hamming_distance
    )
    if trim:
        optimal_weight = dag.trim_optimal_weight(edge_weight_func=weight_func)
    else:
        optimal_weight = dag.optimal_weight_annotate(edge_weight_func=weight_func)
    return optimal_weight


def disambiguate(
    tree,
    compute_cvs=True,
    random_state=None,
    remove_cvs=False,
    adj_dist=False,
    transition_model=parsimony_utils.default_nt_transitions,
    min_ambiguities=False,
    use_internal_node_sequences=False,
):
    """Randomly resolve ambiguous bases using a two-pass Sankoff Algorithm on
    subtrees of consecutive ambiguity codes.

    Args:
        compute_cvs: If true, compute upward sankoff cost vectors. If ``sankoff_upward`` was
            already run on the tree, this may be skipped.
        random_state: A ``random`` module random state, returned by ``random.getstate()``. Output
            from this function is otherwise deterministic.
        remove_cvs: Remove sankoff cost vectors from tree nodes after disambiguation.
        adj_dist: Recompute hamming parsimony distances on tree after disambiguation, and store them
            in ``dist`` node attributes.
        min_ambiguities: If True, leaves ambiguities in reconstructed sequences, expressing which
            bases are possible at each site in a maximally parsimonious disambiguation of the given
            topology. In the history DAG paper, this is known as a strictly min-weight ambiguous labeling.
            Otherwise, sequences are resolved in one possible way, and include no ambiguities.
    """
    if random_state is None:
        random.seed(tree.write(format=1))
    else:
        random.setstate(random_state)

    seq_len = len(next(tree.iter_leaves()).sequence)
    adj_arr = transition_model.get_adjacency_array(seq_len)
    if compute_cvs:
        sankoff_upward(
            tree,
            len(tree.sequence),
            transition_model=transition_model,
            use_internal_node_sequences=use_internal_node_sequences,
        )
    # Second pass of Sankoff: choose bases
    preorder = list(tree.traverse(strategy="preorder"))
    for node in preorder:
        if min_ambiguities:
            adj_vec = node.cost_vector != np.stack(
                (node.cost_vector.min(axis=1),) * len(transition_model.bases), axis=1
            )
            new_seq = [
                transition_model.get_ambiguity_from_tuple(tuple(map(float, row)))
                for row in adj_vec
            ]
        else:
            base_indices = []
            for idx in range(seq_len):
                min_cost = min(node.cost_vector[idx])
                base_index = random.choice(
                    [
                        i
                        for i, val in enumerate(node.cost_vector[idx])
                        if val == min_cost
                    ]
                )
                base_indices.append(base_index)

            adj_vec = adj_arr[np.arange(seq_len), base_indices]
            new_seq = [
                transition_model.bases[base_index] for base_index in base_indices
            ]

        # Adjust child cost vectors
        for child in node.children:
            child.cost_vector += adj_vec
        node.sequence = "".join(new_seq)

    if remove_cvs:
        for node in tree.traverse():
            try:
                node.del_feature("cost_vector")
            except (AttributeError, KeyError):
                pass
    if adj_dist:
        tree.dist = 0
        for node in tree.iter_descendants():
            node.dist = transition_model.weighted_hamming_distance(
                node.up.sequence, node.sequence
            )
    return tree


def build_tree(
    newickstring,
    fasta_map,
    newickformat=1,
    reference_id=None,
    reference_sequence=None,
    ignore_internal_sequences=False,
):
    """Load an ete tree from a newick string, and add 'sequence' attributes
    from fasta.

    If internal node sequences aren't specified in the newick string and fasta data,
    internal node sequences will be fully ambiguous (contain repeated N's).

    Arguments:
        newickstring: a newick string
        fasta_map: a dictionary with sequence id keys matching node names in `newickstring`, and sequence values.
        newickformat: the ete format identifier for the passed newick string. See ete docs
        reference_id: key in `fasta_map` corresponding to the root sequence of the tree, if root sequence is fixed.
        reference_sequence: fixed root sequence of the tree, if root sequence is fixed
        ignore_internal_sequences: if True, sequences at non-leaf nodes specified in fasta_map
        and newickstring will be ignored, and all internal sequences will be fully ambiguous.
    """
    tree = ete3.Tree(newickstring, format=newickformat)
    # all fasta entries should be same length
    seq_len = len(next(iter(fasta_map.values())))
    ambig_seq = "?" * seq_len
    for node in tree.traverse():
        if node.is_root() and reference_sequence is not None:
            node.add_feature("sequence", reference_sequence)
        elif node.is_root() and reference_id is not None:
            node.add_feature("sequence", fasta_map[reference_id])
        elif (not node.is_leaf()) and ignore_internal_sequences:
            node.add_feature("sequence", ambig_seq)
        elif node.name in fasta_map:
            node.add_feature("sequence", fasta_map[node.name])
        else:
            node.add_feature("sequence", ambig_seq)
    return tree


def build_trees_from_files(newickfiles, fastafile, **kwargs):
    """Same as `build_tree`, but takes a list of filenames containing newick
    strings, and a filename for a fasta file, and returns a generator on
    trees."""
    fasta_map = load_fasta(fastafile)
    for newick in newickfiles:
        with open(newick, "r") as fh:
            newick = fh.read()
        yield build_tree(newick, fasta_map, **kwargs)


def parsimony_score(tree, transition_model=parsimony_utils.default_nt_transitions):
    """Returns the parsimony score of a (disambiguated) ete tree.

    Tree must have 'sequence' attributes on all nodes.
    """
    return sum(
        transition_model.weighted_hamming_distance(node.up.sequence, node.sequence)
        for node in tree.iter_descendants()
    )


def remove_invariant_sites(fasta_map):
    """Eliminate invariant characters in a fasta alignment and record the
    0-based indices of those variant sites."""
    # eliminate characters for which there's no diversity:
    informative_sites = [
        idx for idx, chars in enumerate(zip(*fasta_map.values())) if len(set(chars)) > 1
    ]
    newfasta = {
        key: "".join(oldseq[idx] for idx in informative_sites)
        for key, oldseq in fasta_map.items()
    }
    return newfasta, informative_sites


def parsimony_scores_from_topologies(
    newicks, fasta_map, gap_as_char=False, remove_invariants=False, **kwargs
):
    """Returns a generator on parsimony scores of trees specified by newick
    strings and fasta. additional keyword arguments are passed to `build_tree`.

    Args:
        newicks: newick strings of trees whose parsimony scores will be computed
        fasta_map: fasta data as a dictionary, as output by ``load_fasta``
        gap_as_char: if True, gap characters `-` will be treated as a fifth character.
            Otherwise, they will be synonymous with `N`'s.
        remove_invariants: if True, removes invariant sites from the provided fasta_map
    """
    if remove_invariants:
        print("removing invariant sites...")
        newfasta = remove_invariant_sites(fasta_map)
        print("...finished removing invariant sites")
    else:
        newfasta = fasta_map
    yield from (
        sankoff_upward(build_tree(newick, newfasta, **kwargs), gap_as_char=gap_as_char)
        for newick in newicks
    )


def parsimony_scores_from_files(treefiles, fastafile, **kwargs):
    """Returns the parsimony scores of trees specified by newick files and a
    fasta file.

    Arguments match `build_trees_from_files`. Additional keyword
    arguments are passed to ``parsimony_scores_from_topologies``.
    """

    def load_newick(npath):
        with open(npath, "r") as fh:
            return fh.read()

    return parsimony_scores_from_topologies(
        (load_newick(path) for path in treefiles), load_fasta(fastafile), **kwargs
    )


def build_dag_from_trees(trees):
    """Build a history DAG from trees containing a `sequence` attribute on all
    nodes.

    unifurcations in the provided trees will be deleted.
    """
    trees = [tree.copy() for tree in trees]
    for tree in trees:
        for node in tree.iter_descendants():
            if len(node.children) == 1:
                node.delete(prevent_nondicotomic=False)
        if len(tree.children) == 1:
            newchild = tree.add_child()
            newchild.add_feature("sequence", tree.sequence)
    return history_dag_from_etes(
        trees,
        ["sequence"],
    )


def disambiguate_history(history):
    """A rather ugly way to disambiguate a history, by converting to ete, using
    the disambiguate method, then converting back to HistoryDag."""
    seq_len = len(next(history.get_leaves()).label.sequence)
    ambig_seq = "N" * seq_len
    etetree = history.to_ete(
        feature_funcs={
            "sequence": (lambda n: n.label.sequence if n.is_leaf() else ambig_seq)
        }
    )
    disambiguate(etetree, min_ambiguities=True)
    rhistory = history_dag_from_etes([etetree], ["sequence"])
    return rhistory


def treewise_sankoff_in_dag(dag, cover_edges=False):
    """Perform tree-wise sankoff to compute labels for all nodes in the DAG."""
    newdag = history_dag_from_histories(
        disambiguate_history(history)
        for history in dag.iter_covering_histories(cover_edges=cover_edges)
    )
    newdag.explode_nodes()
    newdag.make_complete()
    newdag.trim_optimal_weight()
    newdag.convert_to_collapsed()
    return newdag



from historydag.dag import EdgeSet, from_tree
from test_historydag import newickstring2, newickstring3
import ete3

""" EdgeSet Tests """


def test_init():
    tree = ete3.Tree(newickstring2, format=1)
    dag = from_tree(tree, ["sequence"])
    tree1 = ete3.Tree(newickstring2, format=1)
    dag1 = from_tree(tree1, ["sequence"])
    EdgeSet()
    e = EdgeSet([dag.dagroot])
    for edge in e:
        print(e)
    EdgeSet([dag.dagroot], probs=[0.5])
    try:
        EdgeSet([dag.dagroot, dag1.dagroot])
        raise TypeError("EdgeSet init is allowing identical dags to be added")
    except (TypeError, ValueError):
        pass
    try:
        EdgeSet([dag.dagroot], [dag1.dagroot])
        raise TypeError("EdgeSet init is allowing multiple positional arguments")
    except TypeError:
        pass


def test_iter():
    tree = ete3.Tree(newickstring2, format=1)
    dag = from_tree(tree.children[0], ["sequence"])
    tree1 = ete3.Tree(newickstring3, format=1)
    dag1 = from_tree(tree1.children[0], ["sequence"])
    e2 = EdgeSet([next(dag.dagroot.children()), next(dag1.dagroot.children())])
    for target, weight, prob in e2:
        pass
    for node in (dag | dag1).preorder():
        for clade, eset in node.clades.items():
            assert len(eset.targets) == len(eset.probs) and len(eset.probs) == len(
                eset.weights
            )


def test_copy():
    tree = ete3.Tree(newickstring2, format=1)
    dag = from_tree(tree.children[0], ["sequence"])
    tree1 = ete3.Tree(newickstring3, format=1)
    dag1 = from_tree(tree1.children[0], ["sequence"])
    e2 = EdgeSet([next(dag.dagroot.children()), next(dag1.dagroot.children())])
    e2.shallowcopy()


def test_not_to_root():
    tree = ete3.Tree(newickstring2, format=1)
    dag = from_tree(tree.children[0], ["sequence"])
    tree1 = ete3.Tree(newickstring3, format=1)
    dag1 = from_tree(tree1.children[0], ["sequence"])
    e1 = EdgeSet([next(dag.dagroot.children()), next(dag1.dagroot.children())])
    try:
        EdgeSet([dag.dagroot, next(dag1.dagroot.children())])
        raise ValueError("Adding a root node to an edge set was allowed")
    except ValueError:
        pass
    try:
        e1.add_to_edgeset(dag.dagroot)
        raise ValueError("Adding a root node to an edge set was allowed")
    except ValueError:
        pass



import ete3
import pickle
import historydag
import historydag.dag as hdag
import historydag.utils as dagutils
from historydag.utils import logsumexp
from collections import Counter, namedtuple
import random
from historydag import parsimony_utils
from math import exp, isclose


def normalize_counts(counter):
    n = len(list(counter.elements()))
    return ([num / n for _, num in counter.items()], (n / len(counter)) / n)


def is_close(f1, f2, tol=0.000001):
    return isclose(f1, f2, rel_tol=tol)


def deterministic_newick(tree: ete3.TreeNode) -> str:
    """For use in comparing ete3 TreeNodes with newick strings"""
    newtree = tree.copy()
    for node in newtree.traverse():
        node.name = 1
        node.children.sort(key=lambda node: node.sequence)
        node.dist = 1
    return newtree.write(format=1, features=["sequence"], format_root_node=True)


def deterministic_newick_topology(tree: ete3.TreeNode) -> str:
    """For use in comparing ete3 TreeNodes with newick strings, distinguishing only
    by topologies above leaves."""
    newtree = tree.copy()
    for node in newtree.traverse():
        node.name = node.sequence
        node.children.sort(
            key=lambda node: str(sorted(lf.name for lf in node.get_leaves()))
        )
        node.dist = 1
    return newtree.write(format=9)


newicklistlist = [
    ["((AA, CT)CG, (TA, CC)CG)CC;", "((AA, CT)CA, (TA, CC)CC)CC;"],
    [
        "((CA, GG)CA, AA, (TT, (CC, GA)CC)CC)AA;",
        "((CA, GG)CA, AA, (TT, (CC, GA)CA)CA)AA;",
        "((CA, GG)CG, AA, (TT, (CC, GA)GC)GC)AG;",
    ],
    ["((AA, CT)CG, (TA, CC)CG)CC;", "((AA, CT)CA, (TA, CC)CC)CC;"],
    [
        "((CA, GG)CA, AT, (TT, (CC, GA)CC)CC)AA;",
        "((CA, GG)CA, AA, (TT, (CC, GA)CA)CA)AA;",
        "((CA, GG)CG, AA, (TT, (CC, GA)GC)GC)AG;",
    ],
]

dags = [
    hdag.history_dag_from_newicks(
        newicklist, [], label_functions={"sequence": lambda n: n.name}
    )
    for newicklist in newicklistlist
]

with open("sample_data/toy_trees_100_uncollapsed.p", "rb") as fh:
    uncollapsed = pickle.load(fh)
for tree in uncollapsed:
    if len(tree.children) == 1:
        newchild = tree.copy()
        for child in newchild.get_children():
            newchild.remove_child(child)
        tree.add_child(newchild)
        assert newchild.is_leaf()

dags.append(
    hdag.history_dag_from_etes(
        uncollapsed[0:5], [], label_functions={"sequence": lambda n: n.sequence}
    )
)

with open("tests/smalltestdag.p", "rb") as fh:
    dags.append(pickle.load(fh))

compdags = [dag.copy() for dag in dags]
for dag in compdags:
    dag.make_complete()
dags.extend(compdags)

cdags = [dag.copy() for dag in dags]
for dag in cdags:
    dag.convert_to_collapsed()


def _testfactory(resultfunc, verify_func, collapse_invariant=False, accum_func=Counter):
    for dag, cdag in zip(dags, cdags):
        # check dags
        result = resultfunc(dag)
        verify_result = accum_func([verify_func(tree) for tree in dag.get_histories()])
        assert result == verify_result

        # check cdags
        cresult = resultfunc(cdag)
        cverify_result = accum_func(
            [verify_func(tree) for tree in cdag.get_histories()]
        )
        assert cresult == cverify_result

        # check they agree, if collapse_invariant.
        if collapse_invariant:
            assert result == cresult


def test_valid_dags():
    for dag in dags + cdags:
        dag._check_valid()
        dag.copy()._check_valid()
        # each edge is allowed:
        for node in dag.postorder():
            for clade in node.clades:
                for target in node.clades[clade].targets:
                    assert target.clade_union() == clade or node.is_ua_node()

        # each clade has a descendant edge:
        for node in dag.postorder():
            for clade in node.clades:
                assert len(node.clades[clade].targets) > 0

        # leaf labels are unique:
        leaf_labels = [node.label for node in dag.postorder() if len(node.clades) == 0]
        assert len(set(leaf_labels)) == len(leaf_labels)


def test_count_topologies():
    dagutils.make_newickcountfuncs(internal_labels=False)
    for dag in dags:
        checkset = {
            tree.to_newick(
                name_func=lambda n: n.label.sequence if n.is_leaf() else "",
                features=[],
                feature_funcs={},
            )
            for tree in dag.get_histories()
        }
        print(checkset)
        assert dag.count_topologies_fast() == len(checkset)


def test_count_topologies_equals_newicks():
    for dag in dags:
        assert dag.count_topologies_fast() == dag.count_topologies()


def test_parsimony():
    def parsimony(tree):
        tree.recompute_parents()
        return sum(
            parsimony_utils.hamming_edge_weight(list(node.parents)[0], node)
            for node in tree.postorder()
            if node.parents
        )

    _testfactory(lambda dag: dag.weight_count(), parsimony)


def test_copy():
    # Copying the DAG gives the same DAG back, or at least a DAG expressing
    # the same trees
    _testfactory(
        lambda dag: Counter(tree.to_newick() for tree in dag.copy().get_histories()),
        lambda tree: tree.to_newick(),
    )


def test_newicks():
    # See that the to_newicks method agrees with to_newick applied to all trees in DAG.
    kwargs = {"name_func": lambda n: n.label.sequence, "features": []}
    _testfactory(
        lambda dag: Counter(dag.to_newicks(**kwargs)),
        lambda tree: tree.to_newick(**kwargs),
    )
    kwargs = {"name_func": lambda n: n.label.sequence, "features": ["sequence"]}
    _testfactory(
        lambda dag: Counter(dag.to_newicks(**kwargs)),
        lambda tree: tree.to_newick(**kwargs),
    )
    kwargs = {"name_func": lambda n: "1", "features": ["sequence"]}
    _testfactory(
        lambda dag: Counter(dag.to_newicks(**kwargs)),
        lambda tree: tree.to_newick(**kwargs),
    )
    kwargs = {"name_func": lambda n: "1", "features": None}
    _testfactory(
        lambda dag: Counter(dag.to_newicks(**kwargs)),
        lambda tree: tree.to_newick(**kwargs),
    )
    kwargs = {"name_func": lambda n: "1", "features": []}
    _testfactory(
        lambda dag: Counter(dag.to_newicks(**kwargs)),
        lambda tree: tree.to_newick(**kwargs),
    )


def test_verify_newicks():
    # See that the newick string output is the same as given by ete3
    kwargs = {"name_func": lambda n: n.label.sequence, "features": ["sequence"]}
    invkwargs = {"label_features": ["sequence"], "label_functions": {}}

    def verify(tree):
        etetree = tree.to_ete(**kwargs)
        cladetree = hdag.from_tree(etetree, **invkwargs)
        return cladetree.to_newick(**kwargs)

    _testfactory(lambda dag: Counter(dag.to_newicks(**kwargs)), verify)


def test_collapsed_counts():
    def uncollapsed(tree):
        # Returns the number of uncollapsed edges in the tree
        etetree = tree.to_ete(features=["sequence"])
        return sum(n.up.sequence == n.sequence for n in etetree.iter_descendants())

    _testfactory(
        lambda dag: dag.weight_count(
            edge_weight_func=dagutils.access_nodefield_default("sequence", False)(
                lambda s1, s2: s1 == s2
            )
        ),
        uncollapsed,
    )


def test_min_weight():
    def parsimony(tree):
        tree.recompute_parents()
        return sum(
            parsimony_utils.hamming_edge_weight(list(node.parents)[0], node)
            for node in tree.postorder()
            if node.parents
        )

    _testfactory(
        lambda dag: dag.optimal_weight_annotate(),
        parsimony,
        accum_func=min,
        collapse_invariant=True,
    )


def test_count_histories():
    _testfactory(lambda dag: dag.count_histories(), lambda tree: 1, accum_func=sum)


def test_count_histories_expanded():
    for dag in dags + cdags:
        ndag = dag.copy()
        ndag.explode_nodes()
        assert (
            dag.count_histories(
                expand_func=parsimony_utils.default_nt_transitions.ambiguity_map.get_sequence_resolution_func(
                    "sequence"
                )
            )
            == ndag.count_histories()
        )


def test_count_weights_expanded():
    for dag in dags + cdags:
        ndag = historydag.sequence_dag.SequenceHistoryDag.from_history_dag(dag.copy())
        odag = ndag.copy()
        ndag.explode_nodes()
        assert odag.hamming_parsimony_count() == ndag.weight_counts_with_ambiguities()


def test_topology_decompose():
    # make sure that trimming to a topology results in a DAG expressing exactly
    # the trees which have that topology.
    for collapse_leaves in [False, True]:
        kwargs = dagutils.make_newickcountfuncs(
            internal_labels=False, collapse_leaves=collapse_leaves
        )
        for dag in [dag.copy() for dag in dags]:
            nl = dag.weight_count(**kwargs)
            for idx, (topology, count) in enumerate(nl.items()):
                # print(topology, count, idx)
                trimdag = dag.copy()
                print(trimdag.weight_count(**kwargs))
                print(topology)
                trimdag.trim_topology(topology, collapse_leaves=collapse_leaves)
                assert trimdag.weight_count(**kwargs) == {topology: count}


def test_topology_count_collapse():
    dag = dags[0].copy()
    print(
        dag.weight_count(
            **dagutils.make_newickcountfuncs(
                internal_labels=False, collapse_leaves=True
            )
        )
    )
    assert dag.count_topologies(collapse_leaves=True) == 2


# this tests if each of the trees indexed are valid trees
# they should have exactly one edge descending from each node clade pair
def test_valid_histories():
    for history_dag in dags + cdags:
        newicks = history_dag.to_newicks()
        n_histories = len(history_dag)
        for curr_dag_index in range(0, n_histories):
            if n_histories < 10 or curr_dag_index % 10 == 0:
                next_tree = history_dag[curr_dag_index]
                assert next_tree.to_newick() in newicks
                assert next_tree.is_history()


# this should check if the indexing algorithm accurately
# captures all possible histories of the dag
def test_indexing_comprehensive():
    for history_dag in dags + cdags:
        assert Counter(history_dag.to_newicks()) == Counter(
            history_dag[i].to_newick() for i in range(len(history_dag))
        )


def test_trim_fixedleaves():
    for dag in dags + cdags:
        reftree = dag.sample()
        kwarglist = [
            (dagutils.make_rfdistance_countfuncs(reftree, rooted=True), min),
            (dagutils.make_rfdistance_countfuncs(reftree, rooted=False), min),
        ]
        for kwargs, opt_func in kwarglist:
            dag = dag.copy()
            dag.make_complete()
            dag._check_valid()
            dag.recompute_parents()
            dag._check_valid()
            all_weights = dag.weight_count(**kwargs)
            optimal_weight = dag.trim_optimal_weight(**kwargs, optimal_func=opt_func)
            assert all_weights[optimal_weight] == dag.count_trees()
            dag._check_valid()
            dag.convert_to_collapsed()
            dag._check_valid()


def test_trim():
    kwarglist = [
        (parsimony_utils.hamming_distance_countfuncs, min),
        (dagutils.node_countfuncs, min),
    ]
    for dag in dags + cdags:
        for kwargs, opt_func in kwarglist:
            dag = dag.copy()
            dag.make_complete()
            dag._check_valid()
            dag.recompute_parents()
            dag._check_valid()
            all_weights = dag.weight_count(**kwargs)
            optimal_count = dag.count_optimal_histories(**kwargs, optimal_func=opt_func)
            optimal_weight = dag.trim_optimal_weight(**kwargs, optimal_func=opt_func)
            assert all_weights[optimal_weight] == dag.count_trees() == optimal_count
            dag._check_valid()
            dag.convert_to_collapsed()
            dag._check_valid()


def test_from_nodes():
    for dag in dags + cdags:
        cdag = dag.copy()
        cdag.make_complete()
        cdag.trim_optimal_weight()
        cdag._check_valid()
        wc = cdag.weight_count()
        ndag = hdag.history_dag_from_nodes(cdag.preorder())
        ndag._check_valid()
        ndag.trim_optimal_weight()
        ndag._check_valid()
        print(ndag.to_graphviz())
        assert wc == ndag.weight_count()


def test_sample_with_node():
    random.seed(1)
    dag = dags[-1]
    dag.make_uniform()
    node_to_count = dag.count_nodes()
    min_count = min(node_to_count.values())
    least_supported_nodes = [
        node for node, val in node_to_count.items() if val == min_count
    ]
    for node in least_supported_nodes:
        tree_samples = [dag.sample_with_node(node) for _ in range(min_count * 6)]
        tree_samples[0]._check_valid()
        tree_newicks = {tree.to_newick() for tree in tree_samples}
        # We sampled all trees possible containing the node
        assert len(tree_newicks) == min_count
        # All trees sampled contained the node
        assert all(node in set(tree.preorder()) for tree in tree_samples)
        # # trees containing the node were sampled uniformly
        # # (This is slow but seems to work)
        # norms, avg = normalize_counts(Counter(tree.to_newick() for tree in tree_samples))
        # print(norms)
        # assert all(is_close(norm, avg) for norm in norms)


def test_sample_with_edge():
    random.seed(1)
    dag = dags[-1]
    dag.recompute_parents()
    node_to_count = dag.count_nodes()
    min_count = min(node_to_count.values())
    least_supported_nodes = [
        node for node, val in node_to_count.items() if val == min_count
    ]
    node = least_supported_nodes[0]

    def edges(dag):
        eset = set()
        for node in dag.preorder():
            for child in node.children():
                eset.add((node, child))
        return eset

    for parent in node.parents:
        edge = (parent, node)
        tree_samples = [dag.sample_with_edge(edge) for _ in range(min_count * 5)]
        tree_samples[0]._check_valid()
        # We sampled all trees possible containing the node
        # All trees sampled contained the node
        assert all(edge in edges(tree) for tree in tree_samples)


def test_iter_covering_histories():
    for dag in dags + cdags:
        codag = dag.copy()
        codag.make_complete()
        trees = list(dag.iter_covering_histories())
        tdag = trees[0] | trees
        tdag.make_complete()
        assert tdag.weight_count() == codag.weight_count()


def test_iter_covering_histories_edges():
    for dag in dags + cdags:
        trees = list(dag.iter_covering_histories(cover_edges=True))
        tdag = trees[0] | trees
        assert tdag.weight_count() == dag.weight_count()
        assert set(tdag.to_newicks()) == set(dag.to_newicks())


def test_relabel():
    dag = dags[-1]
    Label = namedtuple("Label", ["sequence", "newthing"])
    ndag = dag.relabel(lambda n: Label(n.label.sequence, len(list(n.children()))))
    Label = namedtuple("Label", ["sequence"])
    odag = ndag.relabel(lambda n: Label(n.label.sequence))
    odag._check_valid()
    assert dag.weight_count() == odag.weight_count()


def test_add_label_fields():
    dag = dags[-1]
    old_fieldset = dag.get_label_type()._fields + tuple(["isLeaf", "originalLocation"])
    new_field_values = {n: [n.is_leaf(), "new location"] for n in dag.postorder()}
    ndag = dag.add_label_fields(
        ["isLeaf", "originalLocation"], lambda n: new_field_values[n]
    )
    ndag._check_valid()
    new_fieldset = ndag.get_label_type()._fields
    assert old_fieldset == new_fieldset


def test_remove_label_fields():
    dag = dags[-1]
    old_fieldset = dag.get_label_type()._fields
    new_field_values = {n: [n.is_leaf()] for n in dag.postorder()}
    ndag = dag.add_label_fields(["added_field"], lambda n: new_field_values[n])
    ndag._check_valid()
    # test removing a field that isn't there and one that is
    odag = dag.remove_label_fields(["removed_field", "added_field"])
    new_fieldset = odag.get_label_type()._fields
    odag._check_valid()
    assert old_fieldset == new_fieldset


# ############# RF Distance Tests: ###############
def rooted_rf_distance(history1, history2):
    cladeset1 = {n.clade_union() for n in history1.preorder(skip_ua_node=True)}
    cladeset2 = {n.clade_union() for n in history2.preorder(skip_ua_node=True)}
    return len(cladeset1 ^ cladeset2)


def test_right_left_rf_add_correctly():
    # In both the rooted and unrooted cases, left and right RF distances should
    # sum to the normal RF distance.
    for rooted in (True, False):
        for dag in dags:
            ref_tree = dag.sample()
            left_kwargs = dagutils.make_rfdistance_countfuncs(
                ref_tree, rooted=rooted, one_sided="left"
            )
            right_kwargs = dagutils.make_rfdistance_countfuncs(
                ref_tree, rooted=rooted, one_sided="right"
            )
            kwargs = dagutils.make_rfdistance_countfuncs(ref_tree, rooted=rooted)

            for tree in dag:
                assert tree.optimal_weight_annotate(
                    **left_kwargs
                ) + tree.optimal_weight_annotate(
                    **right_kwargs
                ) == tree.optimal_weight_annotate(
                    **kwargs
                )


def test_right_left_rf_collapse():
    """
    When one tree is a resolution of another, one-sided RF distance should be
    able to detect this with a distance of 0. The relevant descriptions from
    the docstring:

        one_sided: May be 'left', 'right', or None. 'left' means that we count
            splits (or clades, in the rooted case) which are in the reference tree but not
            in the DAG tree, especially useful if trees in the DAG might be resolutions of
            a multifurcating reference. 'right' means that we count splits or clades in
            the DAG tree which are not in the reference tree, useful if the reference tree
            is possibly a resolution of multifurcating trees in the DAG. If not None,
            one_sided_coefficients are ignored.
    """
    for rooted in (True, False):
        count = 0
        for dag in dags:
            for tree in dag:
                ctree = tree.copy()
                ctree.convert_to_collapsed()
                kwargs = dagutils.make_rfdistance_countfuncs(ctree, rooted=rooted)
                if tree.optimal_weight_annotate(**kwargs) == 0:
                    # Then they're the same topology (when unrooted, a simple
                    # node count isn't enough to identify this)
                    continue
                else:
                    count += 1
                    left_kwargs = dagutils.make_rfdistance_countfuncs(
                        ctree, rooted=rooted, one_sided="left"
                    )
                    assert tree.optimal_weight_annotate(**left_kwargs) == 0
                    oleft_kwargs = dagutils.make_rfdistance_countfuncs(
                        tree, rooted=rooted, one_sided="left"
                    )
                    assert ctree.optimal_weight_annotate(**oleft_kwargs) > 0
                    right_kwargs = dagutils.make_rfdistance_countfuncs(
                        ctree, rooted=rooted, one_sided="right"
                    )
                    assert tree.optimal_weight_annotate(**right_kwargs) > 0
                    oright_kwargs = dagutils.make_rfdistance_countfuncs(
                        tree, rooted=rooted, one_sided="right"
                    )
                    assert ctree.optimal_weight_annotate(**oright_kwargs) == 0
        assert count > 0


def test_rf_rooted_distances():
    for dag in dags:
        ref_tree = dag.sample()
        weight_kwargs = dagutils.make_rfdistance_countfuncs(ref_tree, rooted=True)

        weight_to_self = ref_tree.optimal_weight_annotate(**weight_kwargs)
        if not (weight_to_self == 0):
            print("nonzero distance to self in this tree ^^: ", weight_to_self)
            assert False

        def rf_distance(intree):
            return rooted_rf_distance(intree, ref_tree)

        assert Counter(rf_distance(tree) for tree in dag) == dag.weight_count(
            **weight_kwargs
        )


def test_rf_unrooted_distances():
    for dag in reversed(dags):
        ref_tree = dag.sample()
        weight_kwargs = dagutils.make_rfdistance_countfuncs(ref_tree, rooted=False)
        ref_tree_ete = ref_tree.to_ete(features=["sequence"])
        for node in ref_tree_ete.traverse():
            node.name = node.sequence
        ref_taxa = {n.sequence for n in ref_tree_ete.get_leaves()}
        weight_to_self = ref_tree.optimal_weight_annotate(**weight_kwargs)
        if not (weight_to_self == 0):
            print(ref_tree_ete)
            print("nonzero distance to self in this tree ^^: ", weight_to_self)
            assert False

        def rf_distance(intree):
            intreeete = intree.to_ete(features=["sequence"])
            assert len(intreeete.children) != 1
            return ref_tree_ete.robinson_foulds(
                intreeete, attr_t1="sequence", attr_t2="sequence", unrooted_trees=True
            )[0]

        if Counter(rf_distance(tree) for tree in dag) != dag.weight_count(
            **weight_kwargs
        ):
            print("label format ", next(dag.postorder()).label)
            for tree in dag:
                thistree = tree.to_ete(features=["sequence"])
                tree_taxa = {n.sequence for n in thistree.get_leaves()}
                if tree_taxa != ref_taxa:
                    continue
                ref_dist = rf_distance(tree)
                comp_dist = tree.optimal_weight_annotate(**weight_kwargs)
                if ref_dist != comp_dist:
                    for node in thistree.get_leaves():
                        node.name = node.sequence
                        # node.name = str(namedict[node.sequence])
                    for node in ref_tree_ete.get_leaves():
                        node.name = node.sequence
                        # node.name = str(namedict[node.sequence])
                    print("reference tree:")
                    print(ref_tree_ete)
                    print("this tree:")
                    print(thistree)
                    print("correct RF: ", ref_dist)
                    print("computed RF: ", comp_dist)
                    assert False


def test_optimal_sum_rf_distance():
    # Can only use unrooted sum RF distances on dags containing trees all on
    # the same taxon set.

    def one_taxon_set(dag):
        return len({n.clade_union() for n in dag.dagroot.children()}) == 1

    dags_to_test = [dag for dag in dags if one_taxon_set(dag)]
    assert len(dags_to_test) > 5

    def other(side):
        return {"right": "left", "left": "right", None: None}[side]

    for dag_idx, ref_dag in enumerate(dags_to_test):
        print("dagnum ", dag_idx)
        # let's just do this test for three trees in each dag:
        for tree_idx, tree in zip(range(3), ref_dag):
            for one_sided in ("left", "right", None):
                for rooted in (True, False):
                    print("treenum ", tree_idx)
                    print("one_side ", one_sided)
                    print("rooted ", rooted)
                    # First let's just make sure that when the ref_dag is just a single
                    # tree, optimal_sum_rf_distance agrees with normal rf_distance.
                    single_tree_dag = ref_dag[0]
                    # Here we get all the distances between trees in 'single_tree_dag' and the
                    # reference tree 'tree' (there's only one, since 'single_tree_dag'
                    # only contains one tree:
                    expected = single_tree_dag.count_rf_distances(
                        tree, rooted=rooted, one_sided=one_sided
                    )
                    expected_sum = sum(expected.elements())
                    calculated_sum = tree.optimal_sum_rf_distance(
                        single_tree_dag, rooted=rooted, one_sided=other(one_sided)
                    )
                    assert calculated_sum == expected_sum

                    # Now let's try computing the summed rf distance on tree relative
                    # to ref_dag...

                    # Here we get all the distances between trees in 'dag' and the
                    # reference tree 'tree':
                    expected = ref_dag.count_rf_distances(
                        tree, rooted=rooted, one_sided=one_sided
                    )
                    # Here we sum all elements in the counter, with multiplicity:
                    # in other words we sum all distances from trees in 'dag' to 'tree'
                    expected_sum = sum(expected.elements())
                    # This should calculate the sum RF distance from 'tree' to all
                    # trees in 'dag':
                    calculated_sum = tree.optimal_sum_rf_distance(
                        ref_dag, rooted=rooted, one_sided=other(one_sided)
                    )
                    assert calculated_sum == expected_sum


# ############# END RF Distance Tests: ###############


def test_trim_range():
    for curr_dag in [dags[-1], cdags[-1]]:
        history_dag = curr_dag.copy()
        print("history dag contains ", history_dag.count_trees(), " trees")
        counter = history_dag.weight_count()
        max_weight_passed = list(counter.keys())[int(len(counter.keys()) / 2)]
        min_weight_passed = list(counter.keys())[int(len(counter.keys()) / 2)] - 1

        print("weight count before trimming:")
        print(counter)
        print("max weight passed in:")
        print(max_weight_passed)
        trees_to_merge = [
            tree.copy()
            for tree in history_dag
            if (
                tree.optimal_weight_annotate() <= max_weight_passed
                and tree.optimal_weight_annotate() >= min_weight_passed
            )
        ]
        true_subdag = hdag.history_dag_from_clade_trees(trees_to_merge)

        history_dag.trim_within_range(
            min_weight=min_weight_passed, max_weight=max_weight_passed
        )
        print("weight count after trimming:")
        print(history_dag.weight_count())
        print("weight count expected:")
        print(true_subdag.weight_count())

        difference = set(history_dag.to_newicks()) - set(true_subdag.to_newicks())
        print(
            "number of subtrees included in tree after trimming but not after merging: "
            + str(len(difference))
        )
        difference2 = set(true_subdag.to_newicks()) - set(history_dag.to_newicks())
        print(
            "number of subtrees included in tree after merging but not trimming: "
            + str(len(difference2))
        )
        assert set(true_subdag.to_newicks()) == set(history_dag.to_newicks())


def test_trim_filter():
    dag = dags[-1].copy()
    pars_d = parsimony_utils.hamming_distance_countfuncs
    node_d = hdag.utils.node_countfuncs

    # >>
    cdag = dag.copy()
    pars_f = hdag.utils.HistoryDagFilter(pars_d, min)
    node_f = hdag.utils.HistoryDagFilter(node_d, max)
    p_n_f = pars_f + node_f

    assert (cdag[pars_f][node_f]).weight_count(**p_n_f) == (cdag[p_n_f]).weight_count(
        **p_n_f
    )
    print("sequential trimming is the same as trimming with sum of filters")
    # >>
    rf_d = hdag.utils.make_rfdistance_countfuncs(dag[25])
    rf_f = hdag.utils.HistoryDagFilter(rf_d, min)

    pnf_f = p_n_f + rf_f
    assert (cdag[p_n_f][rf_f]).weight_count(**pnf_f) == (cdag[pnf_f]).weight_count(
        **pnf_f
    )
    print("sequential addition of Filters works")
    # >>
    lc_d = pnf_f.weight_funcs.linear_combination([1, 2, 0])
    lc_f = hdag.utils.HistoryDagFilter(lc_d, min)

    min_weight = cdag.optimal_weight_annotate(**lc_f)
    weights = cdag.weight_count(**pnf_f)
    min_weights = Counter(
        w
        for w in weights.elements()
        if sum(c * ww for c, ww in zip([1, 2, 0], w)) == min_weight
    )
    assert (cdag[lc_f]).weight_count(**pnf_f) == min_weights
    print("linear combination trimming works")

    # >>
    print(node_d)
    print(node_f)
    print(p_n_f)
    print(pnf_f)
    print(lc_f)


def test_weight_range_annotate():
    kwargs = parsimony_utils.hamming_distance_countfuncs
    for dag in dags:
        assert dag.weight_range_annotate(**kwargs) == (
            dag.optimal_weight_annotate(**kwargs, optimal_func=min),
            dag.optimal_weight_annotate(**kwargs, optimal_func=max),
        )


def test_sum_all_pair_rf_distance():
    dag = dags[-1]

    small_dag_1 = dag[0] | (dag[i] for i in range(1, 7))
    small_dag_2 = dag[-1] | (dag[i] for i in range(60, 67))
    small_dag_1.summary()
    small_dag_2.summary()
    for rooted in (False, True):
        for one_sided in ("left", "right", None):
            # check 0 on single-tree dag vs itself:
            assert dag[0].sum_rf_distances(rooted=rooted, one_sided=one_sided) == 0
            assert (
                dag[0].sum_rf_distances(
                    reference_dag=dag[0], rooted=rooted, one_sided=one_sided
                )
                == 0
            )

            # check matches single rf distance between two single-tree dags:
            udag = dag.unlabel()
            assert udag[0].sum_rf_distances(
                reference_dag=udag[-1], rooted=rooted, one_sided=one_sided
            ) == udag[0].optimal_rf_distance(
                udag[-1], rooted=rooted, one_sided=one_sided
            )

            # check matches truth on whole DAG vs self:
            assert dag.sum_rf_distances(rooted=rooted, one_sided=one_sided) == sum(
                dag.count_sum_rf_distances(
                    dag, rooted=rooted, one_sided=one_sided
                ).elements()
            )

            # check matches truth on dag1 vs dag2
            assert small_dag_1.sum_rf_distances(
                reference_dag=small_dag_2, rooted=rooted, one_sided=one_sided
            ) == sum(
                small_dag_1.count_sum_rf_distances(
                    small_dag_2, rooted=rooted, one_sided=one_sided
                ).elements()
            )


def test_sum_weight():
    dag = dags[-1].copy()
    correct = sum(dag.weight_count().elements())
    assert correct == dag.sum_weights()


def test_intersection():
    dag = dags[-1]
    dag1 = hdag.history_dag_from_histories(dag.sample() for _ in range(8)) | dag[0]
    dag2 = hdag.history_dag_from_histories(dag.sample() for _ in range(8)) | dag[0]
    idag = dag1 & dag2

    assert len(idag | dag1) == len(dag1)
    assert len(idag | dag2) == len(dag2)


def compare_decomposed_probabilities(dag, edge_func, log_probabilities):
    if log_probabilities:
        kwargs = {
            "accum_func": sum,
            "start_func": lambda n: 0,
            "edge_weight_func": edge_func,
            "optimal_func": logsumexp,
        }
    else:
        kwargs = {
            "accum_func": dagutils.prod,
            "start_func": lambda n: 1,
            "edge_weight_func": edge_func,
            "optimal_func": sum,
        }

    true_unnormalized_probs = [tree.optimal_weight_annotate(**kwargs) for tree in dag]
    if log_probabilities:
        normalization_constant = logsumexp(true_unnormalized_probs)

        def normalize(prob):
            return prob - normalization_constant

    else:
        normalization_constant = sum(true_unnormalized_probs)

        def normalize(prob):
            return prob / normalization_constant

    dag.probability_annotate(edge_func, log_probabilities=log_probabilities)
    conditional_edge_probs = dag.export_edge_probabilities()
    check_kwargs = kwargs.copy()
    check_kwargs["edge_weight_func"] = lambda n1, n2: conditional_edge_probs[(n1, n2)]
    normalized_product_conditional_probs = [
        tree.optimal_weight_annotate(**check_kwargs) for tree in dag
    ]
    for true_prob, check_prob in zip(
        true_unnormalized_probs, normalized_product_conditional_probs
    ):
        assert is_close(normalize(true_prob), check_prob, tol=0.00001)


def test_conditional_edge_probabilities():
    # first compare to uniform:
    compare_decomposed_probabilities(dag, lambda n1, n2: 1, log_probabilities=False)
    # now try log version for uniform:
    compare_decomposed_probabilities(dag, lambda n1, n2: 0, log_probabilities=True)
    # hamming parsimony weighted support:
    compare_decomposed_probabilities(
        dag,
        lambda n1, n2: exp(-3 * parsimony_utils.hamming_edge_weight(n1, n2)),
        log_probabilities=False,
    )
    # now hamming parsimony log version:
    compare_decomposed_probabilities(
        dag,
        lambda n1, n2: -2 * parsimony_utils.hamming_edge_weight(n1, n2),
        log_probabilities=True,
    )
    # and some other examples...
    compare_decomposed_probabilities(dag, lambda n1, n2: 2, log_probabilities=False)


def test_node_support():
    dag = dags[-1].copy()

    # first compare to uniform:
    dag.probability_annotate(lambda n1, n2: 1, log_probabilities=False)
    nd = dag.node_probabilities(log_probabilities=False)
    od = dag.count_nodes()
    hists = dag.count_histories()
    od = {n: count / hists for n, count in od.items()}
    for node in dag.postorder():
        assert is_close(nd[node], od[node], tol=0.0001)

    # now try log version for uniform:
    dag.probability_annotate(lambda n1, n2: 0, log_probabilities=True)
    nd = dag.node_probabilities(log_probabilities=True)
    print(nd[dag.dagroot])
    for node in dag.postorder():
        assert is_close(exp(nd[node]), od[node], tol=0.0001)


def test_conditional_annotation_nilpotent():
    dag = dags[-1].copy()
    dag.probability_annotate(lambda n1, n2: 1, log_probabilities=False)
    edge_weights = dag.export_edge_probabilities()
    dag.probability_annotate(
        lambda n1, n2: edge_weights[(n1, n2)], log_probabilities=False
    )
    comp_edge_weights = dag.export_edge_probabilities()
    print(len(comp_edge_weights))
    for key in edge_weights:
        assert isclose(edge_weights[key], comp_edge_weights[key])


def test_count_nodes():
    dag = dags[-1].copy()
    node_counts = dag.count_nodes()
    dag.uniform_distribution_annotate(log_probabilities=False)
    node_supports = dag.node_probabilities(log_probabilities=False)
    n_histories = dag.count_histories()
    for node in node_counts:
        assert node_counts[node] == round(node_supports[node] * n_histories)

    edge_supports = dag.edge_probabilities(log_probabilities=False)
    edge_counts = dag.count_edges()
    for edge in edge_counts:
        assert edge_counts[edge] == round(edge_supports[edge] * n_histories)

    # now with some collapsing
    dag = dags[-1].copy()
    node_counts = dag.count_nodes(collapse=True)
    dag.uniform_distribution_annotate(log_probabilities=False)
    node_supports = dag.node_probabilities(
        log_probabilities=False, collapse_key=hdag.HistoryDagNode.clade_union
    )
    n_histories = dag.count_histories()
    for node in node_counts:
        assert node_counts[node] == round(node_supports[node] * n_histories)

    edge_supports = dag.edge_probabilities(
        log_probabilities=False,
        collapse_key=lambda edge: (edge[0].clade_union(), edge[1].clade_union()),
    )
    edge_counts = dag.count_edges(collapsed=True)
    for edge in edge_counts:
        assert edge_counts[edge] == round(edge_supports[edge] * n_histories)

    # Now counting splits:
    dag = dags[-1].copy()

    def history_to_splits(history):
        splits = set()
        all_taxa = next(history.dagroot.children()).clade_union()
        for node in history.preorder(skip_ua_node=True):
            node_clade = node.clade_union()
            splits.add(frozenset({node_clade, all_taxa - node_clade}))
        return splits - frozenset({all_taxa, frozenset()})

    split_sets = [history_to_splits(history) for history in dag]
    split_counts = dag.count_nodes(collapse=True, rooted=False)
    for split, count in split_counts.items():
        assert count == sum(1 for s in split_sets if split in s)


def test_likelihoods():
    dag = dags[-1]
    assert dag.optimal_weight_annotate(**historydag.likelihoods.JC_log_countfuncs) < 0


def test_in_keyword():
    assert dags[-1][0] in dags[-1]
    assert dags[-3][0] not in dags[-1]


def test_bifurcating_count():
    for dag in dags:
        assert dag.count_histories(bifurcating=True) >= dag.count_histories(
            bifurcating=False
        )


def test_ascii_visualization():
    dag = dags[-1]
    history = dag[0]
    print(history.to_ascii(lambda n: hash(n)))
    print(history.to_ascii(lambda n: hash(n), sort_method="ladderize"))
    print(history.to_ascii(lambda n: hash(n), sort_method="child-name"))
    print(history.to_ascii(lambda n: hash(n), sort_method="leaf-name"))

    hdag.ascii_compare_histories(
        history, dag[1], lambda n: hash(n), sort_method="child-name"
    )
    hdag.ascii_compare_histories(history.dagroot, dag[1].dagroot, lambda n: hash(n))



from math import exp
import ete3
import random
from typing import (
    List,
    Callable,
    Any,
    Mapping,
    Generator,
    Iterable,
    Set,
    Optional,
    Tuple,
    Dict,
    FrozenSet,
)
from copy import deepcopy
from historydag import utils
from historydag.utils import Weight, Label, UALabel


class HistoryDagNode:
    r"""A recursive representation of a history DAG object.

    - a dictionary keyed by clades (frozensets) containing EdgeSet objects
    - a label, which is a namedtuple.
    """

    def __init__(self, label: Label, clades: dict, attr: Any):
        self.clades = clades
        # If passed a nonempty dictionary, need to add self to children's parents
        self.label = label
        self.parents: Set[HistoryDagNode] = set()
        self.attr = attr
        self._dp_data: Any = None
        if self.clades:
            for _, edgeset in self.clades.items():
                edgeset.parent = self
            for child in self.children():
                child.parents.add(self)

        if len(self.clades) == 1:
            raise ValueError(
                "Internal nodes (those which are not the DAG UA root node) "
                "may not have exactly one child clade; Unifurcations cannot be expressed "
                "in the history DAG."
            )

    def __repr__(self) -> str:
        return str((self.label, set(self.clades.keys())))

    def __hash__(self) -> int:
        return hash((self.label, self.child_clades()))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, HistoryDagNode):
            return (self.label, self.child_clades()) == (
                other.label,
                other.child_clades(),
            )
        else:
            raise NotImplementedError

    def __le__(self, other: object) -> bool:
        if isinstance(other, HistoryDagNode):
            return (self.label, self.sorted_child_clades()) <= (
                other.label,
                other.sorted_child_clades(),
            )
        else:
            raise NotImplementedError

    def __lt__(self, other: object) -> bool:
        if isinstance(other, HistoryDagNode):
            return (self.label, self.sorted_child_clades()) < (
                other.label,
                other.sorted_child_clades(),
            )
        else:
            raise NotImplementedError

    def __gt__(self, other: object) -> bool:
        return not self.__le__(other)

    def __ge__(self, other: object) -> bool:
        return not self.__lt__(other)

    def empty_copy(self) -> "HistoryDagNode":
        """Returns a HistoryDagNode object with the same clades, label, and
        attr dictionary, but no descendant edges."""
        return HistoryDagNode(
            self.label, {clade: EdgeSet() for clade in self.clades}, deepcopy(self.attr)
        )

    def node_self(self) -> "HistoryDagNode":
        """Deprecated name for :meth:`empty_copy`"""
        return self.empty_copy()

    def clade_union(self) -> FrozenSet[Label]:
        r"""Returns the union of this node's child clades (or a set containing
        only the node label, for leaf nodes.)"""
        if self.is_leaf():
            return frozenset([self.label])
        else:
            return frozenset().union(*self.clades.keys())

    def under_clade(self) -> FrozenSet[Label]:
        """Deprecated name for :meth:`clade_union`"""
        return self.clade_union()

    def is_leaf(self) -> bool:
        """Returns whether this is a leaf node."""
        return not bool(self.clades)

    def is_ua_node(self) -> bool:
        """Returns whether this is the source node in the DAG, from which all
        others are reachable."""
        return False

    def is_root(self) -> bool:
        """Deprecated name for :meth:`is_ua_node`"""
        return self.is_ua_node()

    def is_history_root(self) -> bool:
        """Return whether node is a root of any histories in the DAG."""
        return any(n.is_ua_node() for n in self.parents)

    def child_clades(self) -> frozenset:
        """Returns the node's child clades, or a frozenset containing a
        frozenset if this node is a UANode."""
        return frozenset(self.clades.keys())

    def partitions(self) -> frozenset:
        """Deprecated name for :meth:`child_clades`"""
        return self.child_clades()

    def sorted_child_clades(self) -> tuple:
        """Returns the node's child clades as a sorted tuple containing leaf
        labels in sorted tuples."""
        return tuple(sorted([tuple(sorted(clade)) for clade in self.clades.keys()]))

    def sorted_partitions(self) -> tuple:
        """Deprecated name for :meth:`sorted_child_clades`"""
        return self.sorted_child_clades()

    def children(
        self, clade: Set[Label] = None
    ) -> Generator["HistoryDagNode", None, None]:
        r"""Returns generator object containing child nodes.

        Args:
            clade: If clade is provided, returns generator object of edge targets from that
        clade. If no clade is provided, generator includes all children of self.
        """
        if clade is None:
            return (
                target for clade in self.clades for target, _, _ in self.clades[clade]
            )
        else:
            return (child for child, _, _ in self.clades[clade])

    def add_edge(
        self,
        target: "HistoryDagNode",
        weight: Weight = 0,
        prob: float = None,
        prob_norm: bool = True,
    ) -> bool:
        r"""Adds edge, if allowed and not already present.

        Returns whether edge was added.
        """
        # target clades must union to a clade of self
        key = frozenset() if self.is_ua_node() else target.clade_union()
        if key not in self.clades:
            raise KeyError("Target clades' union is not a clade of this parent node: ")
        else:
            target.parents.add(self)
            return self.clades[key].add_to_edgeset(
                target,
                weight=weight,
                prob=prob,
                prob_norm=prob_norm,
            )

    def _get_subhistory_by_subid(self, subid: int) -> "HistoryDagNode":
        r"""Returns the subtree below the current HistoryDagNode corresponding
        to the given index."""
        if self.is_leaf():  # base case - the node is a leaf
            return self
        else:
            history = self.empty_copy()

            # get the subtree for each of the clades
            for clade, eset in self.clades.items():
                # get the sum of subtrees of the edges for this clade
                num_subtrees = 0
                for child, weight, _ in eset:
                    num_subtrees = num_subtrees + child._dp_data
                curr_index = subid % num_subtrees

                # find the edge corresponding to the curr_index
                for child, weight, _ in eset:
                    if curr_index >= child._dp_data:
                        curr_index = curr_index - child._dp_data
                    else:
                        # add this edge to the tree somehow
                        history.clades[clade].add_to_edgeset(
                            child._get_subhistory_by_subid(curr_index)
                        )
                        break

                subid = subid / num_subtrees
        return history

    def remove_edge_by_clade_and_id(self, target: "HistoryDagNode", clade: frozenset):
        key: frozenset
        if self.is_ua_node():
            key = frozenset()
        else:
            key = clade
        self.clades[key].remove_from_edgeset_byid(target)

    def remove_node(self, nodedict: Dict["HistoryDagNode", "HistoryDagNode"] = {}):
        r"""Recursively removes node self and any orphaned children from dag.

        May not work on root. Does not check to make sure that parent
        clade still has descendant edges.
        """
        if self in nodedict:
            nodedict.pop(self)
        for child in self.children():
            if self in child.parents:
                child.parents.remove(self)
            if not child.parents:
                child.remove_node(nodedict=nodedict)
        for parent in self.parents:
            parent.remove_edge_by_clade_and_id(self, self.clade_union())
        self.removed = True

    def to_ascii(
        self,
        name_func,
        show_internal=False,
        compact=False,
        sort_method=None,
    ):
        """A convenience function that uses the :meth:`to_ete` method and
        ete3's ASCII drawing tools to render a sub-history below self.

        Expects that this node is part of a tree-shaped history DAG. If this is not
        true, this method may fail silently.

        Args:
            name_func: A function taking a HistoryDagNode and returning a string
                to identify that node in the ascii tree.
            show_internal: Whether to show internal node names.
            compact: Whether to show the tree in a more compact format
            sort_method: Either None, `ladderize`, `leaf-name`, or `child-name`.
                `leaf-name` sorts children by the alphabetically first leaf name
                below each child node. `child-name` sorts directly by child name.
        Returns:
            A string including whitespace and newlines (no tabs) which when printed
            shows the structure of the history.
        """
        t = self.to_ete(name_func=name_func)

        def child_name_sort(tree):
            for node in tree.traverse(strategy="postorder"):
                node.children.sort(key=lambda n: n.name)

        def leaf_name_sort(tree):
            for node in tree.traverse(strategy="postorder"):
                if node.is_leaf():
                    node.firstleaf = node.name
                else:
                    node.children.sort(key=lambda n: n.firstleaf)
                    node.firstleaf = node.children[0].firstleaf

        try:
            sort_func = {
                None: (lambda tree: None),
                "ladderize": t.ladderize,
                "leaf-name": leaf_name_sort,
                "child-name": child_name_sort,
            }[sort_method]
        except KeyError:
            raise KeyError(
                "to_ascii method accepts sort_method None, `ladderize`, "
                f"`child-name`, or `leaf-name`, not {sort_method}."
            )
        sort_func(t)
        return t.get_ascii(show_internal=show_internal, compact=compact)

    def to_ete(
        self,
        name_func: Callable[["HistoryDagNode"], str] = lambda n: "unnamed",
        feature_funcs: Mapping[str, Callable[["HistoryDagNode"], str]] = {},
        sort_func=lambda seq: seq,
    ) -> ete3.TreeNode:
        """Convert a history DAG node which is part of a history to an ete
        tree.

        Args:
            name_func: A map from nodes to newick node names
            feature_funcs: A dictionary keyed by extended newick field names, containing
                functions specifying how to populate that field for each node.

        Returns:
            An ete3 Tree with the same topology as the subhistory below self,
            and node names and attributes as specified.
        """
        node = ete3.TreeNode()
        node.name = name_func(self)
        for feature, func in feature_funcs.items():
            node.add_feature(feature, func(self))
        for child in self.children():
            node.add_child(child.to_ete(name_func, feature_funcs, sort_func))
        return node

    def _sample(
        self, edge_selector=lambda n: True, log_probabilities=False
    ) -> "HistoryDagNode":
        r"""Samples a history (a sub-history DAG containing the root and all
        leaf nodes).

        Returns a new HistoryDagNode object.
        """
        sample = self.empty_copy()
        for clade, eset in self.clades.items():
            mask = [edge_selector((self, target)) for target in eset.targets]
            sampled_target, target_weight = eset.sample(
                mask=mask, log_probabilities=log_probabilities
            )
            sampled_target_subsample = sampled_target._sample(
                edge_selector=edge_selector,
                log_probabilities=log_probabilities,
            )
            sampled_target_subsample.parents = set([self])
            sample.clades[clade].add_to_edgeset(
                sampled_target_subsample,
                weight=target_weight,
            )
        return sample

    def _get_subhistories(self) -> Generator["HistoryDagNode", None, None]:
        r"""Return a generator to iterate through all trees expressed by the
        DAG."""

        def genexp_func(clade):
            # Return generator expression of all possible choices of tree
            # structure from dag below clade
            def f():
                eset = self.clades[clade]
                return (
                    (clade, targettree, i)
                    for i, target in enumerate(eset.targets)
                    for targettree in target._get_subhistories()
                )

            return f

        optionlist = [genexp_func(clade) for clade in self.clades]

        # TODO is this duplicated code?
        for option in utils.cartesian_product(optionlist):
            tree = self.empty_copy()
            for clade, targettree, index in option:
                tree.clades[clade].add_to_edgeset(
                    targettree,
                    weight=self.clades[clade].weights[index],
                )
            yield tree

    def _newick_label(
        self,
        name_func: Callable[["HistoryDagNode"], str] = (lambda n: "unnamed"),
        features: Iterable[str] = None,
        feature_funcs: Mapping[str, Callable[["HistoryDagNode"], str]] = {},
    ) -> str:
        """Return an extended newick format node label.

        Args:
            name_func: A function which maps nodes to names
            features: A list of label fields to be recorded in extended newick format
            feature_funcs: A dictionary keyed by extended newick field names containing
                functions which map nodes to field values. These override fields named
                in `features`, if a key in `feature_funcs` is also contained in `features`.

        Returns:
            A string which can be used as a node name in a newick string.
            For example, `namefuncresult[&&NHX:feature1=val1:feature2=val2]`
        """
        if self.is_ua_node():
            return self.label  # type: ignore
        else:
            if features is None:
                features = self.label._fields
            # Use dict to avoid duplicate fields
            nameval_dict = {
                name: val
                for name, val in self.label._asdict().items()
                if name in features
            }
            nameval_dict.update({name: f(self) for name, f in feature_funcs.items()})
            featurestr = ":".join(f"{name}={val}" for name, val in nameval_dict.items())
            return name_func(self) + (f"[&&NHX:{featurestr}]" if featurestr else "")


class UANode(HistoryDagNode):
    r"""A universal ancestor node, the root node of a HistoryDag."""

    def __init__(self, targetnodes: "EdgeSet"):
        self.label = UALabel()
        # an empty frozenset is not used as a key in any other node
        self.targetnodes = targetnodes
        self.clades = {frozenset(): targetnodes}
        self.parents = set()
        self.attr = dict()
        for child in self.children():
            child.parents.add(self)

    def empty_copy(self) -> "UANode":
        """Returns a UANode object with the same clades and label, but no
        descendant edges."""
        newnode = UANode(EdgeSet())
        newnode.attr = deepcopy(self.attr)
        return newnode

    def is_ua_node(self) -> bool:
        """Returns whether this is the source node in the DAG, from which all
        others are reachable."""
        return True


class EdgeSet:
    r"""A container class for edge target nodes, and associated probabilities
    and weights.

    Goal: associate targets (edges) with arbitrary parameters, but support
    set-like operations like lookup and enforce that elements are unique.
    """

    def __init__(
        self,
        *args,
        weights: Optional[List[float]] = None,
        probs: Optional[List[float]] = None,
    ):
        r"""Takes no arguments, or an ordered iterable containing target
        nodes."""
        if len(args) == 0:
            targets = []
        elif len(args) == 1:
            targets = args[0]
        else:
            raise TypeError(
                f"__init__() takes 0 or 1 positional arguments but {len(args)} were given."
            )
        self.set_targets(targets, weights, probs)

    def __iter__(self):
        return (
            (self.targets[i], self.weights[i], self.probs[i])
            for i in range(len(self.targets))
        )

    def set_targets(self, targets, weights=None, probs=None):
        """Set the target nodes of this node.

        If no weights or probabilities are provided, then these will be
        set to 0 and 1/n, respectively.
        """
        n = len(targets)
        if len(set(targets)) != n:
            raise ValueError(
                f"duplicate target nodes provided: {len(set(targets))} out of {len(targets)} unique."
            )

        self.targets = targets
        self._targetset = set(targets)
        if weights is None:
            weights = [0] * n
        if probs is None:
            if n == 0:
                probs = []
            else:
                probs = [float(1) / n] * n
        self.set_edge_stats(weights, probs)

    def set_edge_stats(self, weights=None, probs=None):
        """Set the edge weights and/or probabilities of this EdgeSet."""
        n = len(self.targets)
        if weights is not None:
            if len(weights) != n:
                raise ValueError(
                    "length of provided weights list must match number of target nodes"
                )
            self.weights = weights
        if probs is not None:
            if len(probs) != n:
                raise ValueError(
                    "length of provided probabilities list must match number of target nodes"
                )
            self.probs = probs

    def shallowcopy(self) -> "EdgeSet":
        """Return an identical EdgeSet object, which points to the same target
        nodes."""
        return EdgeSet(
            [node for node in self.targets],
            weights=self.weights.copy(),
            probs=self.probs.copy(),
        )

    def remove_from_edgeset_byid(self, target_node):
        idlist = [id(target) for target in self.targets]
        if id(target_node) in idlist:
            idx_to_remove = idlist.index(id(target_node))
            self.targets.pop(idx_to_remove)
            self.probs.pop(idx_to_remove)
            self.weights.pop(idx_to_remove)
            self._targetset = set(self.targets)

    def sample(
        self, mask=None, log_probabilities=False
    ) -> Tuple[HistoryDagNode, float]:
        """Returns a randomly sampled child edge, and its corresponding weight.

        When possible, only edges pointing to child nodes on which
        ``selection_function`` evaluates to True will be sampled.
        """
        if log_probabilities:
            weights = [exp(weight) for weight in self.probs]
        else:
            weights = self.probs

        if mask is None or sum(mask) == 0:
            pass
        else:
            weights = [factor * prob for factor, prob in zip(mask, weights)]

        index = random.choices(list(range(len(self.targets))), weights=weights, k=1)[0]
        return (self.targets[index], self.weights[index])

    def add_to_edgeset(self, target, weight=0, prob=None, prob_norm=True) -> bool:
        """Add a target node to the EdgeSet.

        currently does nothing if edge is already present. Also does nothing
        if the target node has one child clade, and parent node is not the DAG root.

        Args:
            target: target node
            weight: edge weight
            prob: edge probability. If not provided, edge probability will be
                1 / n where n is the number of edges in the edgeset.
            prob_norm: if True, probability vector will be normalized.

        Returns:
            Whether an edge was added
        """
        if target.is_ua_node():
            raise ValueError(
                "Edges that target UA nodes are not allowed. "
                f"Target node has label {target.label} and therefore "
                "is assumed to be the DAG UA root node."
            )
        elif target in self._targetset:
            return False
        else:
            self._targetset.add(target)
            self.targets.append(target)
            self.weights.append(weight)

            if prob is None:
                prob = 1.0 / len(self.targets)
            if prob_norm:
                self.probs = list(
                    map(lambda x: x * (1 - prob) / sum(self.probs), self.probs)
                )
            self.probs.append(prob)
            return True


def empty_node(
    label: Label, clades: Iterable[FrozenSet[Label]], attr: Any = None
) -> HistoryDagNode:
    """Return a HistoryDagNode with the given label and clades, with no
    children."""
    return HistoryDagNode(label, {clade: EdgeSet() for clade in clades}, attr)



import ete3
import historydag.dag as hdag
from historydag import utils, parsimony_utils
import pickle
from test_factory import deterministic_newick

newickstring3 = (
    "((4[&&NHX:name=4:sequence=K],(6[&&NHX:name=6:sequence=J],"
    "7[&&NHX:name=7:sequence=I])5[&&NHX:name=5:sequence=H])2[&&NHX:name=2:sequence=H],"
    "8[&&NHX:name=8:sequence=F],(11[&&NHX:name=11:sequence=E],"
    "10[&&NHX:name=10:sequence=D])9[&&NHX:name=9:sequence=C], 12[&&NHX:name=9:sequence=Z])"
    "3[&&NHX:name=3:sequence=H];"
)


def collapse_adjacent_sequences(tree: ete3.TreeNode) -> ete3.TreeNode:
    """Collapse nonleaf nodes that have the same sequence"""
    # Need to keep doing this until the tree fully collapsed. See gctree for this!
    tree = tree.copy()
    to_delete = []
    for node in tree.get_descendants():
        # This must stay invariably hamming distance, since it's measuring equality of strings
        if (
            not node.is_leaf()
            and parsimony_utils.default_nt_transitions.weighted_hamming_distance(
                node.up.sequence, node.sequence
            )
            == 0
        ):
            to_delete.append(node)
    for node in to_delete:
        node.delete()
    return tree


etetree = list(
    hdag.history_dag_from_etes(
        [ete3.TreeNode(newick=newickstring3, format=1)], ["sequence"]
    ).get_histories()
)[0].to_ete(features=["sequence"])
etetree2 = utils.collapse_adjacent_sequences(etetree.copy())


with open("sample_data/toy_trees_100_collapsed.p", "rb") as fh:
    collapsed = pickle.load(fh)
with open("sample_data/toy_trees_100_uncollapsed.p", "rb") as fh:
    uncollapsed = pickle.load(fh)
trees = collapsed + uncollapsed
for tree in trees:
    if len(tree.children) == 1:
        newchild = tree.copy()
        for child in newchild.get_children():
            newchild.remove_child(child)
        tree.add_child(newchild)
        assert newchild.is_leaf()


def test_fulltree():
    dag = hdag.history_dag_from_etes([etetree], ["sequence"])
    dag.convert_to_collapsed()
    dag._check_valid()
    assert set(
        deterministic_newick(tree.to_ete()) for tree in dag.get_histories()
    ) == set({deterministic_newick(etetree2)})


def test_twotrees():
    dag = hdag.history_dag_from_etes([etetree, etetree2], ["sequence"])
    dag.convert_to_collapsed()
    dag._check_valid()
    assert dag.count_histories() == 1
    assert {deterministic_newick(tree.to_ete()) for tree in dag.get_histories()} == {
        deterministic_newick(etetree2)
    }


def test_collapse():
    # This may need a second look
    uncollapsed_dag = hdag.history_dag_from_etes(trees, ["sequence"])
    uncollapsed_dag.convert_to_collapsed()
    allcollapsedtrees = [utils.collapse_adjacent_sequences(tree) for tree in trees]
    collapsed_dag = hdag.history_dag_from_etes(allcollapsedtrees, ["sequence"])
    collapsed_dag._check_valid()
    maybecollapsedtrees = [tree.to_ete() for tree in uncollapsed_dag.get_histories()]
    assert all(utils.is_collapsed(tree) for tree in maybecollapsedtrees)
    n_before = uncollapsed_dag.count_histories()
    uncollapsed_dag.merge(collapsed_dag)
    uncollapsed_dag._check_valid()
    assert n_before == uncollapsed_dag.count_histories()


def test_add_allowed_edges():
    # See that adding only edges that preserve parent labels preserves parsimony
    dag = hdag.history_dag_from_etes(trees, ["sequence"])
    dag.make_complete(preserve_parent_labels=True)
    dag._check_valid()
    c = dag.weight_count()
    assert min(c) == max(c)

    # See that adding only edges between nodes with different labels preserves collapse
    allcollapsedtrees = [collapse_adjacent_sequences(tree) for tree in trees]
    collapsed_dag = hdag.history_dag_from_etes(allcollapsedtrees, ["sequence"])
    collapsed_dag._check_valid()
    collapsed_dag.convert_to_collapsed()
    collapsed_dag._check_valid()
    collapsed_dag.make_complete(adjacent_labels=False)
    collapsed_dag._check_valid()
    assert all(
        parent.label != target.label
        for parent in collapsed_dag.postorder()
        for target in parent.children()
        if not target.is_leaf()
    )



import historydag as hdag
import pickle

with open("sample_data/toy_trees.p", "rb") as f:
    trees = pickle.load(f)

dag = hdag.history_dag_from_etes(trees, ["sequence"])
dag.recompute_parents()

seq_len = len(next(dag.postorder()).label.sequence)
new_seq = "A" * seq_len


def test_inserting_leafnode_everywhere():
    dag_with_leaf_added_everywhere = dag.copy()
    dag_with_leaf_added_everywhere.add_node_at_all_possible_places("A" * seq_len)
    dag_with_leaf_added_everywhere.recompute_parents()
    assert dag_with_leaf_added_everywhere._check_valid()

    num_leaf = len([x for x in dag.postorder() if x.is_leaf()])
    num_treeroots = len([x for x in dag.dagroot.children()])
    # expected number of nodes added:
    #   1 for dagroot,
    #   1 for the newly created node,
    #   1 for each leaf,
    #   2+number non-single-leaf clades for each internal nonroot node
    #   1+num_clades for each treeroot node
    expected_num_with_all_additions = (
        2
        + num_leaf
        + sum(
            [
                2
                + len(n.child_clades())
                - len([c for c in n.child_clades() if len(c) < 2])
                for n in dag.postorder()
                if not (n.is_leaf() or n.is_ua_node())
            ]
        )
        - num_treeroots
    )
    assert (
        len(list(dag_with_leaf_added_everywhere.postorder()))
        == expected_num_with_all_additions
    ), "failed to add expected set of new nodes"


def test_inserting_leafnode_at_closest_spot():
    dag_with_leaf_added_closer = dag.copy()
    dag_with_leaf_added_closer.insert_node("A" * seq_len)
    dag_with_leaf_added_closer.recompute_parents()
    assert dag_with_leaf_added_closer._check_valid()

    # check we successfully added one new node
    assert (
        len(list(dag_with_leaf_added_closer.postorder()))
        == len(list(dag.postorder())) + 1
    ), "failed to add one node successfully"



import historydag.parsimony as pars
import historydag.parsimony_utils as pars_utils
import click
import pickle


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
def _cli():
    """A collection of tools for calculating parsimony scores of newick trees,
    and using them to create a history DAG."""
    pass


@_cli.command("remove-invariants")
@click.argument("in_fasta")
@click.argument("out_fasta")
@click.argument("out_variant_sites")
def _cli_remove_invariant_sites(in_fasta, out_fasta, out_variant_sites):
    fasta_map = pars.load_fasta(in_fasta)
    newfasta, variant_sites = pars.remove_invariant_sites(fasta_map)
    with open(out_fasta, "w") as fh:
        for seqid, seq in newfasta.items():
            print(">" + seqid, file=fh)
            print(seq, file=fh)
    with open(out_variant_sites, "w") as fh:
        print(",".join([str(site) for site in variant_sites]), file=fh)


@_cli.command("build-trees")
@click.argument("treefiles", nargs=-1)
@click.option(
    "-f",
    "--fasta-file",
    required=True,
    help="Filename of a fasta file containing sequences appearing on nodes of newick tree",
)
@click.option(
    "-r",
    "--root-id",
    default=None,
    help="The fasta identifier of the fixed root of provided trees. May be omitted if there is no fixed root sequence.",
)
@click.option(
    "-F",
    "--newick-format",
    default=1,
    help=(
        "Newick format of the provided newick file. See "
        "http://etetoolkit.org/docs/latest/reference/reference_tree.html#ete3.TreeNode"
    ),
)
@click.option(
    "-i",
    "--include-internal-sequences",
    is_flag=True,
    help="include non-leaf node labels, and associated sequences in the fasta file.",
)
@click.option(
    "-g",
    "--gap-as-char",
    is_flag=True,
    help="Treat gap character `-` as a fifth character. Otherwise treated as ambiguous `N`.",
)
@click.option(
    "-a",
    "--preserve-ambiguities",
    is_flag=True,
    help=(
        "Do not disambiguate fully, but rather preserve ambiguities to express "
        "all maximally parsimonious assignments at each site."
    ),
)
@click.option(
    "-o", "--outdir", default=".", help="Directory in which to write pickled trees."
)
@click.option(
    "-c",
    "--clean-trees",
    is_flag=True,
    help="remove cost vectors from tree, resulting in smaller pickled tree files",
)
def _cli_build_trees(
    treefiles,
    fasta_file,
    root_id,
    newick_format,
    include_internal_sequences,
    gap_as_char,
    preserve_ambiguities,
    outdir,
    clean_trees,
):
    trees = pars.build_trees_from_files(
        treefiles,
        fasta_file,
        reference_id=root_id,
        ignore_internal_sequences=(not include_internal_sequences),
    )
    if gap_as_char:
        transition_model = pars_utils.default_nt_gaps_transitions
    else:
        transition_model = pars_utils.default_nt_gaps_transitions
    trees = (
        pars.disambiguate(
            tree,
            min_ambiguities=preserve_ambiguities,
            remove_cvs=clean_trees,
            transition_model=transition_model,
        )
        for tree in trees
    )
    for treefile, tree in zip(treefiles, trees):
        print("saving tree")
        with open(outdir + f"tree_{treefile.split('/')[-1]}.p", "wb") as fh:
            fh.write(pickle.dumps(tree))
        print("saved tree")


@_cli.command("parsimony_scores")
@click.argument("treefiles", nargs=-1)
@click.option(
    "-f",
    "--fasta-file",
    required=True,
    help="Filename of a fasta file containing sequences appearing on nodes of newick tree",
)
@click.option(
    "-r",
    "--root-id",
    default=None,
    help="The fasta identifier of the fixed root of provided trees. May be omitted if there is no fixed root sequence.",
)
@click.option(
    "-F",
    "--newick-format",
    default=1,
    help=(
        "Newick format of the provided newick file. See "
        "http://etetoolkit.org/docs/latest/reference/reference_tree.html#ete3.TreeNode"
    ),
)
@click.option(
    "-i",
    "--include-internal-sequences",
    is_flag=True,
    help="include non-leaf node labels, and associated sequences in the fasta file.",
)
@click.option(
    "-d",
    "--save-to-dag",
    default=None,
    help="Combine loaded and disambiguated trees into a history DAG, and save pickled DAG to provided path.",
)
@click.option(
    "-g",
    "--gap-as-char",
    is_flag=True,
    help="Treat gap character `-` as a fifth character. Otherwise treated as ambiguous `N`.",
)
@click.option(
    "-a",
    "--preserve-ambiguities",
    is_flag=True,
    help=(
        "Do not disambiguate fully, but rather preserve ambiguities to express "
        "all maximally parsimonious assignments at each site."
    ),
)
def _cli_parsimony_score_from_files(
    treefiles,
    fasta_file,
    root_id,
    newick_format,
    include_internal_sequences,
    save_to_dag,
    gap_as_char,
    preserve_ambiguities,
):
    """Print the parsimony score of one or more newick files."""
    pscore_gen = pars.parsimony_scores_from_files(
        treefiles,
        fasta_file,
        reference_id=root_id,
        gap_as_char=gap_as_char,
        ignore_internal_sequences=(not include_internal_sequences),
        remove_invariants=True,
    )

    for score, treepath in zip(pscore_gen, treefiles):
        print(treepath)
        print(score)

    if gap_as_char:
        transition_model = pars_utils.default_nt_gaps_transitions
    else:
        transition_model = pars_utils.default_nt_gaps_transitions
    if save_to_dag is not None:
        trees = pars.build_trees_from_files(
            treefiles,
            fasta_file,
            reference_id=root_id,
            ignore_internal_sequences=(not include_internal_sequences),
        )
        trees = (
            pars.disambiguate(
                tree,
                transition_model=transition_model,
                min_ambiguities=preserve_ambiguities,
            )
            for tree in trees
        )
        dag = pars.build_dag_from_trees(trees)
        with open(save_to_dag, "wb") as fh:
            fh.write(pickle.dumps(dag))


@_cli.command("summarize-dag")
@click.argument("dagpath")
def _cli_summarize_dag(dagpath):
    """Print summary information about the provided history DAG."""
    with open(dagpath, "rb") as fh:
        dag = pickle.load(fh)
    pars.summarize_dag(dag)


if __name__ == "__main__":
    _cli()



import ete3
from historydag.mutation_annotated_dag import (
    load_MAD_protobuf_file,
    load_MAD_protobuf,
    load_json_file,
    AmbiguousLeafCGHistoryDag,
)
from historydag.sequence_dag import SequenceHistoryDag
from math import isclose, log
from historydag.parsimony_utils import default_nt_transitions
from historydag.compact_genome import read_alignment
from historydag import from_tree

# pbdag = load_MAD_protobuf_file("sample_data/full_dag.pb")
pbdag = load_MAD_protobuf_file(
    "sample_data/small_test_proto.pb", compact_genomes=True, node_ids=True
)


def test_provided_leaf_seqs():
    reference = pbdag.get_reference_sequence()
    # first one just to make sure we can read an alignment without providing
    # a reference
    read_alignment("sample_data/small_test_proto.fa")

    alignment = read_alignment(
        "sample_data/small_test_proto.fa", reference_sequence=reference
    )
    assert len(alignment) == 43
    odag = load_MAD_protobuf_file(
        "sample_data/small_test_proto.pb",
        compact_genomes=True,
        node_ids=True,
        leaf_cgs=alignment,
    )
    assert set(n.label for n in pbdag.get_leaves()) == set(
        n.label for n in odag.get_leaves()
    )


def test_load_protobuf():
    top_dag = load_MAD_protobuf_file(
        "sample_data/node_id_dag.pb", compact_genomes=False, node_ids=False
    )
    cg_nid_dag = load_MAD_protobuf_file(
        "sample_data/node_id_dag.pb", compact_genomes=True, node_ids=False
    )
    nid_dag = load_MAD_protobuf_file(
        "sample_data/node_id_dag.pb", compact_genomes=False, node_ids=True
    )
    cg_dag = load_MAD_protobuf_file(
        "sample_data/node_id_dag.pb", compact_genomes=True, node_ids=False
    )

    # These values are from Larch, and may not actually be correct
    for dag in [cg_nid_dag, nid_dag, cg_dag]:
        assert dag.count_histories() == 120679531678039887072
        assert dag._check_valid()

    for dag in [cg_nid_dag, cg_dag]:
        assert dag.optimal_weight_annotate() == 428
        assert dag.optimal_weight_annotate(optimal_func=max) == 1008

    ndag = load_MAD_protobuf(cg_nid_dag.to_protobuf())
    ndag._check_valid()
    ndag = load_MAD_protobuf(cg_dag.to_protobuf())
    ndag._check_valid()
    assert top_dag.count_histories() == cg_nid_dag.unlabel().count_histories()


def test_load_ambiguous_protobuf():
    reference_id = "GL"

    alignment = {
        "GL": "AAAAAAAA",
        "seq1": "CAAAGAAG",
        "seq2": "CGAAGATG",
        "seq3": "AGTANNSN",
        "seq4": "AGTAGAAA",
        "seq5": "CATAAAAG",
    }

    reference_seq = alignment[reference_id]

    def make_tree(internal="GL"):
        internal_seq = alignment[internal]
        t = ete3.Tree()
        t.populate(len(alignment), names_library=alignment)
        for n in t.traverse():
            if n.is_leaf():
                n.add_feature("sequence", alignment[n.name])
            else:
                n.add_feature("sequence", internal_seq)
        dag = from_tree(t, ["sequence"], label_functions={"node_id": lambda n: n.name})
        dag = AmbiguousLeafCGHistoryDag.from_history_dag(dag, reference=reference_seq)
        return dag

    dag = make_tree() | make_tree("seq1") | make_tree("seq2")

    pb = dag.to_protobuf(randomize_leaf_muts=True)
    cdag = load_MAD_protobuf(pb, compact_genomes=True)
    assert isinstance(cdag, AmbiguousLeafCGHistoryDag)
    assert cdag.weight_count() == dag.weight_count()


def test_load_json():
    dag = pbdag.copy()
    dag = dag.remove_label_fields(["node_id"])
    dag._check_valid()
    test_filename = "_test_write_json.json"
    dag.to_json_file(test_filename)
    ndag = load_json_file(test_filename)
    assert dag.test_equal(ndag)
    ndag._check_valid()
    assert dag.weight_count() == ndag.weight_count()


def test_weight_count():
    hamming_cg_edge_weight = default_nt_transitions.weighted_cg_hamming_edge_weight(
        "compact_genome",
        count_root_muts=False,
    )
    kwargs = {"edge_weight_func": hamming_cg_edge_weight}
    sdag = SequenceHistoryDag.from_history_dag(pbdag.copy())
    cdag = pbdag.copy()
    assert cdag.weight_count(**kwargs) == sdag.weight_count()
    assert cdag.optimal_weight_annotate(**kwargs) == sdag.optimal_weight_annotate()
    assert cdag.trim_optimal_weight(**kwargs) == sdag.trim_optimal_weight()


def test_adjusted_node_support():
    dag = load_MAD_protobuf_file(
        "sample_data/small_test_proto.pb", compact_genomes=True, node_ids=False
    )
    dag.convert_to_collapsed()
    dag.uniform_distribution_annotate(log_probabilities=False)
    adj_d = dag.adjusted_node_probabilities(log_probabilities=False)
    d = dag.node_probabilities(log_probabilities=False)

    dag.uniform_distribution_annotate(log_probabilities=True)
    adj_d_log = dag.adjusted_node_probabilities(log_probabilities=True)
    d_log = dag.node_probabilities(log_probabilities=True)

    for node in adj_d:
        adj_p = adj_d[node]
        adj_p_log = adj_d_log[node]
        p = d[node]
        p_log = d_log[node]
        assert isclose(log(p), p_log, abs_tol=1e-09)
        assert isclose(log(adj_p), adj_p_log, abs_tol=1e-09)
        assert adj_p <= p



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dag.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\tdag.proto\x12\x08ProtoDAG"M\n\x03mut\x12\x10\n\x08position\x18\x01 \x01(\x05\x12\x0f\n\x07par_nuc\x18\x02 \x01(\x05\x12\x0f\n\x07mut_nuc\x18\x03 \x03(\x05\x12\x12\n\nchromosome\x18\x05 \x01(\t"\x92\x01\n\x04\x65\x64ge\x12\x0f\n\x07\x65\x64ge_id\x18\x01 \x01(\x03\x12\x13\n\x0bparent_node\x18\x02 \x01(\x03\x12\x14\n\x0cparent_clade\x18\x03 \x01(\x03\x12\x12\n\nchild_node\x18\x04 \x01(\x03\x12%\n\x0e\x65\x64ge_mutations\x18\x05 \x03(\x0b\x32\r.ProtoDAG.mut\x12\x13\n\x0b\x65\x64ge_weight\x18\x06 \x01(\x02"6\n\tnode_name\x12\x0f\n\x07node_id\x18\x01 \x01(\x03\x12\x18\n\x10\x63ondensed_leaves\x18\x02 \x03(\t"{\n\x04\x64\x61ta\x12\x1d\n\x05\x65\x64ges\x18\x01 \x03(\x0b\x32\x0e.ProtoDAG.edge\x12\'\n\nnode_names\x18\x02 \x03(\x0b\x32\x13.ProtoDAG.node_name\x12\x14\n\x0creference_id\x18\x03 \x01(\t\x12\x15\n\rreference_seq\x18\x04 \x01(\tb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "dag_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _MUT._serialized_start = 23
    _MUT._serialized_end = 100
    _EDGE._serialized_start = 103
    _EDGE._serialized_end = 249
    _NODE_NAME._serialized_start = 251
    _NODE_NAME._serialized_end = 305
    _DATA._serialized_start = 307
    _DATA._serialized_end = 430
# @@protoc_insertion_point(module_scope)



# Remember to add any additional functions/modules to docs/api.rst
from .dag import (  # noqa
    HistoryDag,
    HistoryDagNode,
    from_tree,
    empty_node,
    from_newick,
    history_dag_from_newicks,
    history_dag_from_etes,
    history_dag_from_histories,
    history_dag_from_trees,
)

from . import _version

from . import (  # noqa
    utils,
    parsimony_utils,
    parsimony,
    mutation_annotated_dag,
    sequence_dag,
    compact_genome,
    likelihoods,
)

try:
    # requires dendropy
    from . import beast_loader  # noqa
except ModuleNotFoundError:
    pass

__version__ = _version.get_versions()["version"]

# patching deprecated call paths here to avoid circular import in utils:
utils.sequence_resolutions_count = utils._deprecate_message(
    "`utils.sequence_resolutions_count` deprecated. Use the `get_sequence_resolution_count_func` method from an "
    "appropriate `parsimony_utils.AmbiguityMap` object`"
)(
    parsimony_utils.standard_nt_ambiguity_map.get_sequence_resolution_count_func(
        "sequence"
    )
)

utils.sequence_resolutions = utils._deprecate_message(
    "`utils.sequence_resolutions` deprecated. Use the `get_sequence_resolution_func` method from an "
    "appropriate `parsimony_utils.AmbiguityMap` object`"
)(parsimony_utils.standard_nt_ambiguity_map.get_sequence_resolution_func("sequence"))



import historydag.parsimony_utils as parsimony_utils
from historydag.utils import AddFuncDict
from math import log


def _JC_branch_length_raw(seq_len, mutated_sites):
    p = mutated_sites / seq_len
    return -(0.75) * log(1 - ((4 * p) / 3))


def _JC_log_edge_weight_raw(seq_len, mutated_sites):
    if mutated_sites == 0:
        return 0
    else:
        N = seq_len
        m = mutated_sites
        return m * (log(m) - log(3 * N)) + (N - m) * (log(N - m) - log(N))


def JC_branch_length(seq1, seq2):
    return _JC_branch_length_raw(
        len(seq1),
        parsimony_utils.default_nt_transitions.weighted_hamming_distance(seq1, seq2),
    )


def JC_log_edge_weight(seq1, seq2):
    mutated_sites = parsimony_utils.default_nt_transitions.weighted_hamming_distance(
        seq1, seq2
    )
    return _JC_log_edge_weight_raw(len(seq1), mutated_sites)


def JC_cg_branch_length(seq1, seq2):
    return _JC_branch_length_raw(
        len(seq1.reference),
        parsimony_utils.default_nt_transitions.weighted_cg_hamming_distance(seq1, seq2),
    )


def JC_cg_log_edge_weight(seq1, seq2):
    mutated_sites = parsimony_utils.default_nt_transitions.weighted_cg_hamming_distance(
        seq1, seq2
    )
    return _JC_log_edge_weight_raw(len(seq1.reference), mutated_sites)


JC_cg_branch_length_countfuncs = AddFuncDict(
    {
        "start_func": lambda n: 0,
        "edge_weight_func": lambda n1, n2: 0
        if n1.is_ua_node()
        else JC_cg_branch_length(n1.label.compact_genome, n2.label.compact_genome),
        "accum_func": sum,
    },
    name="JukesCantorBranchLength",
)

JC_cg_log_countfuncs = AddFuncDict(
    {
        "start_func": lambda n: 0,
        "edge_weight_func": lambda n1, n2: 0
        if n1.is_ua_node()
        else JC_cg_log_edge_weight(n1.label.compact_genome, n2.label.compact_genome),
        "accum_func": sum,
    },
    name="JukesCantorLogLikelihood",
)

JC_branch_length_countfuncs = AddFuncDict(
    {
        "start_func": lambda n: 0,
        "edge_weight_func": lambda n1, n2: 0
        if n1.is_ua_node()
        else JC_branch_length(n1.label.sequence, n2.label.sequence),
        "accum_func": sum,
    },
    name="JukesCantorBranchLength",
)

JC_log_countfuncs = AddFuncDict(
    {
        "start_func": lambda n: 0,
        "edge_weight_func": lambda n1, n2: 0
        if n1.is_ua_node()
        else JC_log_edge_weight(n1.label.sequence, n2.label.sequence),
        "accum_func": sum,
    },
    name="JukesCantorLogLikelihood",
)
